# -*- coding: utf-8 -*-

""" merge data files """

import argparse
import csv
import logging
import os
import sys
import tempfile

from datetime import timedelta
from functools import lru_cache
from pathlib import Path

# pylint: disable=no-name-in-module
from pyspark.sql import SparkSession
from pyspark.sql.functions import array, length, lower, size, to_timestamp, when
from pytility import clear_list, concat_files, parse_int
from scrapy.utils.misc import arg_to_iter

from .utils import now, to_lower

csv.field_size_limit(sys.maxsize)

LOGGER = logging.getLogger(__name__)


@lru_cache()
def _spark_session(log_level=None):
    LOGGER.info("creating Spark session with log level <%s>", log_level)

    os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"

    try:
        builder = (
            SparkSession.builder.appName(__name__)
            .config("spark.ui.showConsoleProgress", False)
            .config("spark.executor.memory", "16G")
            .config("spark.driver.memory", "16G")
            .config("spark.driver.maxResultSize", "16G")
        )

        master = os.getenv("SPARK_MASTER")
        if master:
            LOGGER.info("Using Spark master <%s>", master)
            builder = builder.master(master)

        spark = builder.getOrCreate()

        if log_level:
            spark.sparkContext.setLogLevel(log_level)

        return spark

    except Exception:
        LOGGER.exception("unable to create Spark session")

    return None


def _compare(first, second, column="_latest"):
    return (
        second
        if not first[column] or (second[column] and second[column] >= first[column])
        else first
    )


def _column_type(column, column_type=None):
    column_type = to_lower(column_type)
    return (
        to_timestamp(column)
        if column_type in ("date", "datetime", "dt")
        else lower(column)
        if column_type in ("istr", "istring", "lower")
        else column
    )


def _remove_empty(data, remove_false=False):
    for column, dtype in data.dtypes:
        if dtype in ("string", "binary"):
            LOGGER.info("Remove empty string or binary values from <%s>", column)
            data = data.withColumn(column, when(length(data[column]) > 0, data[column]))
        # TODO recursive in maps and structs
        elif dtype.startswith("array") or dtype.startswith("map"):
            LOGGER.info("Remove empty array or map values from <%s>", column)
            data = data.withColumn(column, when(size(data[column]) > 0, data[column]))
        elif dtype == "boolean" and remove_false:
            LOGGER.info("Remove <false> values from <%s>", column)
            data = data.withColumn(column, when(data[column], data[column]))
    return data


def merge_files(
    in_paths,
    out_path,
    keys="id",
    key_types=None,
    latest=None,
    latest_types=None,
    latest_min=None,
    latest_required=False,
    fieldnames=None,
    fieldnames_exclude=None,
    sort_keys=False,
    sort_latest=False,
    sort_fields=None,
    sort_descending=False,
    concat_output=False,
    log_level=None,
    dry_run: bool = False,
):
    """merge files into one"""

    in_paths = list(map(str, arg_to_iter(in_paths)))

    if dry_run:
        LOGGER.info(
            "[DRY RUN] Merging items from [%d paths] into <%s>",
            len(in_paths),
            out_path,
        )
        for in_path in in_paths:
            LOGGER.info("[DRY RUN] Merging file <%s>", in_path)
        LOGGER.info("[DRY RUN] Done merging.")
        return

    spark = _spark_session(log_level=log_level)

    if spark is None:
        raise RuntimeError(
            "Please make sure Spark is installed and configured correctly!"
        )

    LOGGER.info(
        "Merging items from %s into <%s> with Spark session %r",
        f"[{len(in_paths) } paths]" if len(in_paths) > 10 else in_paths,
        out_path,
        spark,
    )

    fieldnames = clear_list(arg_to_iter(fieldnames))
    fieldnames_exclude = frozenset(arg_to_iter(fieldnames_exclude))

    if fieldnames and fieldnames_exclude:
        LOGGER.warning(
            "Both <fieldnames> and <fieldnames_exclude> were specified, please choose one"
        )

    sort_fields = tuple(arg_to_iter(sort_fields))
    if sum(map(bool, (sort_keys, sort_latest, sort_fields))) > 1:
        LOGGER.warning(
            "Only use at most one of <sort_keys>, <sort_latest>, and <sort_fields>"
        )

    keys = tuple(arg_to_iter(keys))
    key_types = tuple(arg_to_iter(key_types))
    key_types += (None,) * (len(keys) - len(key_types))
    assert len(keys) == len(key_types)
    LOGGER.info("Using keys %s with types %s", keys, key_types)

    latest = tuple(arg_to_iter(latest))
    latest_types = tuple(arg_to_iter(latest_types))
    latest_types += (None,) * (len(latest) - len(latest_types))
    assert len(latest) == len(latest_types)
    LOGGER.info("Using latest %s with types %s", latest, latest_types)

    data = spark.read.json(path=in_paths, mode="DROPMALFORMED", dropFieldIfAllNull=True)

    key_column_names = [f"_key_{i}" for i in range(len(keys))]
    key_columns = [
        _column_type(data[column], column_type).alias(name)
        for column, column_type, name in zip(keys, key_types, key_column_names)
    ]
    key_columns_str = (column.cast("string") for column in key_columns)
    latest_column_names = [f"_latest_{i}" for i in range(len(latest))]
    latest_columns = [
        _column_type(data[column], column_type).alias(name)
        for column, column_type, name in zip(latest, latest_types, latest_column_names)
    ]
    latest_columns_str = (column.cast("string") for column in latest_columns)

    drop_subset = keys + tuple(key_column_names)
    if latest_required:
        drop_subset += latest + tuple(latest_column_names)
    LOGGER.info("Dropping rows without values in columns %s", drop_subset)

    data = data.select(
        "*",
        *key_columns,
        array(*key_columns_str).alias("_key"),
        *latest_columns,
        array(*latest_columns_str).alias("_latest"),
    ).dropna(how="any", subset=drop_subset)

    if latest_min is not None:
        LOGGER.info("Filter out items before %s", latest_min)
        data = data.filter(latest_columns[0] >= latest_min)

    rdd = (
        data.rdd.keyBy(lambda row: tuple(arg_to_iter(row["_key"])))
        .reduceByKey(_compare)
        .values()
    )

    data = rdd.toDF(schema=data.schema)

    if sort_keys:
        LOGGER.info(
            "Sorting %s by columns %s",
            "descending" if sort_descending else "ascending",
            keys,
        )
        data = data.sort(*key_column_names, ascending=not sort_descending)
    elif sort_latest:
        LOGGER.info(
            "Sorting %s by columns %s",
            "descending" if sort_descending else "ascending",
            latest,
        )
        data = data.sort(*latest_column_names, ascending=not sort_descending)
    elif sort_fields:
        LOGGER.info(
            "Sorting %s by columns %s",
            "descending" if sort_descending else "ascending",
            sort_fields,
        )
        data = data.sort(*sort_fields, ascending=not sort_descending)

    data = data.drop("_key", *key_column_names, "_latest", *latest_column_names)

    columns = frozenset(data.columns) - fieldnames_exclude
    if fieldnames:
        fieldnames = [column for column in fieldnames if column in columns]
        LOGGER.info("Only use columns: %s", fieldnames)
    else:
        fieldnames = sorted(columns)
        LOGGER.info("Use sorted column names: %s", fieldnames)
    data = data.select(*fieldnames)

    data = _remove_empty(data)

    if concat_output:
        with tempfile.TemporaryDirectory() as temp_path:
            path = Path(temp_path) / "out"
            LOGGER.info("Saving temporary output to <%s>", path)
            data.write.json(path=str(path))

            LOGGER.info("Concatenate temporary files to <%s>", out_path)
            files = path.glob("part-*")
            concat_files(dst=out_path, srcs=sorted(files), ensure_newline=True)

    else:
        LOGGER.info("Saving output to <%s>", out_path)
        data.write.json(path=str(out_path))

    LOGGER.info("Done merging.")


def _parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("paths", nargs="*", help="input JSON Lines files and dirs")
    parser.add_argument("--out-path", "-o", help="output path")
    parser.add_argument(
        "--keys", "-k", nargs="+", default=("id",), help="target columns for merging"
    )
    parser.add_argument(
        "--key-types",
        "-K",
        nargs="+",
        choices=("str", "string", "istr", "istring", "int", "float", "date"),
        help="target column data type",
    )
    parser.add_argument(
        "--latest", "-l", nargs="+", help="target column for latest item"
    )
    parser.add_argument(
        "--latest-types",
        "-L",
        nargs="+",
        choices=("str", "string", "int", "float", "date"),
        help="latest column data type",
    )
    parser.add_argument(
        "--latest-min",
        "-m",
        help="minimum value for latest column, all other values will be ignored (days for dates)",
    )
    parser.add_argument(
        "--latest-required",
        "-r",
        action="store_true",
        help="drop rows that do not have values in their latest column(s)",
    )
    fields_group = parser.add_mutually_exclusive_group()
    fields_group.add_argument("--fields", "-f", nargs="+", help="output columns")
    fields_group.add_argument(
        "--fields-exclude", "-F", nargs="+", help="ignore these output columns"
    )
    sort_group = parser.add_mutually_exclusive_group()
    sort_group.add_argument(
        "--sort-keys", "-s", action="store_true", help="sort output by keys"
    )
    sort_group.add_argument(
        "--sort-latest",
        "-S",
        action="store_true",
        help='sort output by "latest" column',
    )
    sort_group.add_argument("--sort-fields", nargs="+", help="sort output by columns")
    parser.add_argument(
        "--sort-desc",
        "-D",
        action="store_true",
        help="sort descending",
    )
    parser.add_argument(
        "--concat", "-c", action="store_true", help="concatenate output into one file"
    )
    parser.add_argument("--dry-run", "-n", action="store_true", help="dry run")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    latest_min = (
        now() - timedelta(days=parse_int(args.latest_min))
        if args.latest_min
        and args.latest_types
        and to_lower(args.latest_types[0]) == "date"
        else args.latest_min
    )

    merge_files(
        in_paths=args.paths,
        out_path=args.out_path,
        keys=args.keys,
        key_types=args.key_types,
        latest=args.latest,
        latest_types=args.latest_types,
        latest_min=latest_min,
        latest_required=args.latest_required,
        fieldnames=args.fields,
        fieldnames_exclude=args.fields_exclude,
        sort_keys=args.sort_keys,
        sort_latest=args.sort_latest,
        sort_fields=args.sort_fields,
        sort_descending=args.sort_desc,
        concat_output=args.concat,
        log_level="DEBUG"
        if args.verbose > 1
        else "INFO"
        if args.verbose > 0
        else "WARN",
        dry_run=args.dry_run,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""News syncing, merging, splitting, and uploading."""

import argparse
import logging
import os.path
import sys

from datetime import date, timedelta, timezone
from pathlib import Path
from shutil import rmtree
from subprocess import run
from time import sleep
from typing import TYPE_CHECKING, Optional, Union

from pytility import parse_date

from .merge import merge_files
from .split import split_files
from .utils import date_from_file, now

if TYPE_CHECKING:
    try:
        from git import Repo
    except ImportError:
        pass

LOGGER = logging.getLogger(__name__)
BASE_DIR = Path(__file__).resolve().parent.parent


def _get_git_repo(path: Union[Path, str, None]) -> Optional["Repo"]:
    if not path:
        return None

    path = Path(path).resolve()

    try:
        from git import InvalidGitRepositoryError, NoSuchPathError, Repo
    except ImportError:
        LOGGER.exception("Unable to import Git library")
        return None

    try:
        LOGGER.info("Trying to find Git repo at <%s> or its parents", path)
        repo = Repo(path=path, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        LOGGER.exception("Path <%s> does not point to a valid Git repo", path)
        return None

    return repo


def update_news(
    *,
    s3_src,
    path_feeds,
    path_merged,
    path_split,
    split_git_update=False,
    s3_dst=None,
    split_size=None,
    log_level=None,
    dry_run: bool = False,
):
    """News syncing, merging, splitting, and uploading."""

    dry_run_prefix = "[DRY RUN] " if dry_run else ""

    path_feeds = Path(path_feeds).resolve()
    path_merged = Path(path_merged).resolve()
    path_split = Path(path_split).resolve()

    LOGGER.info(
        "%sSync from <%s>, merge from <%s> into <%s>, split into <%s>",
        dry_run_prefix,
        s3_src,
        path_feeds,
        path_merged,
        path_split,
    )

    if split_git_update:
        repo = _get_git_repo(path_split.parent)
        if repo is None or not repo.working_dir:
            repo = None
            split_git_update = False
            path_git = None
            git_rel_path = None
            LOGGER.error(
                "%sUnable to update Git repo <%s>",
                dry_run_prefix,
                path_split.parent,
            )

        else:
            path_git = Path(repo.working_dir).resolve()
            git_rel_path = path_split.parent.relative_to(path_git)
            LOGGER.info("%sUpdate Git repo <%s>", dry_run_prefix, path_git)
            if not dry_run:
                LOGGER.info("Pulling latest commits from remote <%s>", repo.remotes[0])
                try:
                    repo.remotes[0].pull(ff_only=True)
                except Exception:
                    LOGGER.exception(
                        "Unable to pull from remote <%s> to repo <%s>",
                        repo.remotes[0],
                        path_git,
                    )

    else:
        repo = None

    if s3_dst:
        LOGGER.info("%sUpload results to <%s>", dry_run_prefix, s3_dst)

    LOGGER.info("%sDeleting existing dir <%s>", dry_run_prefix, path_split.parent)
    if not dry_run:
        if repo is None:
            rmtree(path_split.parent, ignore_errors=True)
        else:
            try:
                deleted_files = repo.index.remove(
                    items=[str(git_rel_path)],
                    working_tree=True,
                    r=True,
                )
                LOGGER.info(
                    "Deleted %d files from <%s>",
                    len(deleted_files),
                    path_split.parent,
                )
            except Exception:
                LOGGER.exception(
                    "Unable to remove path <%s> from Git",
                    path_split.parent,
                )

        path_feeds.mkdir(parents=True, exist_ok=True)
        path_merged.parent.mkdir(parents=True, exist_ok=True)
        path_split.parent.mkdir(parents=True, exist_ok=True)

    LOGGER.info("%sS3 sync from <%s> to <%s>", dry_run_prefix, s3_src, path_feeds)
    if not dry_run:
        run(["aws", "s3", "sync", s3_src, os.path.join(path_feeds, "")], check=True)

    merge_files(
        in_paths=path_feeds.rglob("*.jl"),
        out_path=path_merged,
        keys="article_id",
        key_types="string",
        latest=("published_at", "scraped_at"),
        latest_types=("date", "date"),
        latest_required=True,
        sort_latest=True,
        sort_descending=True,
        concat_output=True,
        log_level=log_level,
        dry_run=dry_run,
    )

    split_files(
        path_in=path_merged,
        path_out=path_split,
        size=split_size,
        exclude_empty=True,
        dry_run=dry_run,
    )

    if repo is not None and git_rel_path:
        message = f"Automatic commit by <{__name__}> {date.today().isoformat()}"
        LOGGER.info(
            "%sCommitting changes to repo <%s> with message <%s>",
            dry_run_prefix,
            path_git,
            message,
        )
        if not dry_run:
            try:
                repo.index.add(items=[str(git_rel_path)])
                repo.index.commit(message=message, skip_hooks=True)
            except Exception:
                LOGGER.exception(
                    "There was a problem commit to Git repo <%s>",
                    path_git,
                )

        for remote in repo.remotes:
            LOGGER.info("%sPushing changes to remote <%s>", dry_run_prefix, remote)
            if not dry_run:
                try:
                    remote.push()
                except Exception:
                    LOGGER.exception(
                        "There was a problem pushing to remote <%s>",
                        remote,
                    )

    if s3_dst:
        LOGGER.info(
            "%sS3 sync from <%s> to <%s>",
            dry_run_prefix,
            path_split.parent,
            s3_dst,
        )
        if not dry_run:
            run(
                [
                    "aws",
                    "s3",
                    "sync",
                    "--acl",
                    "public-read",
                    "--exclude",
                    ".gitignore",
                    "--exclude",
                    ".DS_Store",
                    "--exclude",
                    ".bucket",
                    "--size-only",
                    "--delete",
                    os.path.join(path_split.parent, ""),
                    s3_dst,
                ],
                check=True,
            )

    LOGGER.info("%sDone updating news.", dry_run_prefix)


def _parse_args():
    parser = argparse.ArgumentParser(
        description="News syncing, merging, splitting, and uploading.",
    )
    parser.add_argument(
        "--src-bucket",
        "-b",
        default="scrape.news.recommend.games",
        help="S3 bucket with scraped data",
    )
    parser.add_argument(
        "--dst-bucket",
        "-B",
        # default="news.recommend.games",
        help="S3 bucket to upload to",
    )
    parser.add_argument(
        "--feeds",
        "-f",
        default=BASE_DIR / "feeds" / "news",
        help="Scraped items",
    )
    parser.add_argument(
        "--merged",
        "-m",
        default=BASE_DIR / "feeds" / "news_merged.jl",
        help="Merged file",
    )
    parser.add_argument(
        "--split",
        "-s",
        default=BASE_DIR / "feeds" / "news_hosting" / "news_{number:05d}.json",
        help="Split file template",
    )
    parser.add_argument(
        "--split-size",
        "-S",
        type=int,
        default=25,
        help="number of items in each result file",
    )
    parser.add_argument(
        "--git",
        "-g",
        action="store_true",
        help="Update the Git repo of the split files",
    )
    parser.add_argument(
        "--dont-run-before",
        "-d",
        help="Either a date or a file with date information",
    )
    parser.add_argument(
        "--interval",
        "-i",
        type=int,
        default=10 * 60,  # 10 minutes
        help="number of seconds to wait before next run",
    )
    parser.add_argument("--dry-run", "-n", action="store_true", help="dry run")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    dont_run_before = parse_date(
        args.dont_run_before,
        tzinfo=timezone.utc,
    ) or date_from_file(
        args.dont_run_before,
        tzinfo=timezone.utc,
    )

    if dont_run_before:
        LOGGER.info("Don't run before %s", dont_run_before.isoformat())
        sleep_seconds = dont_run_before.timestamp() - now().timestamp()
        if sleep_seconds > 0:
            LOGGER.info("Going to sleep for %.1f seconds", sleep_seconds)
            sleep(sleep_seconds)

    if args.interval and args.dont_run_before and not parse_date(args.dont_run_before):
        dont_run_before = now() + timedelta(seconds=args.interval)
        LOGGER.info(
            "Don't run next time before %s, writing tag to <%s>",
            dont_run_before.isoformat(),
            args.dont_run_before,
        )
        with open(args.dont_run_before, "w", encoding="utf-8") as file_obj:
            file_obj.write(dont_run_before.isoformat())

    update_news(
        s3_src=f"s3://{args.src_bucket}/",
        path_feeds=args.feeds,
        path_merged=args.merged,
        path_split=args.split,
        s3_dst=f"s3://{args.dst_bucket}/" if args.dst_bucket else None,
        split_size=args.split_size,
        split_git_update=args.git,
        log_level="DEBUG"
        if args.verbose > 1
        else "INFO"
        if args.verbose > 0
        else "WARN",
        dry_run=args.dry_run,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

""" util functions """

import json
import logging
import os
import re

from datetime import datetime, timezone
from functools import lru_cache
from pathlib import Path
from types import GeneratorType
from typing import Any, Dict, Iterable, List, Optional, Pattern, Union
from urllib.parse import ParseResult, parse_qs, unquote_plus, urlparse

from pytility import (
    arg_to_iter,
    clear_list,
    normalize_space,
    parse_int,
    take_first,
    to_str,
    parse_date,
)
from scrapy.item import BaseItem, Item
from w3lib.html import replace_entities

try:
    # pylint: disable=redefined-builtin
    from smart_open import open
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)

REGEX_ENTITIES = re.compile(r"(&#(\d+);)+")
REGEX_SINGLE_ENT = re.compile(r"&#(\d+);")

REGEX_BGG_ID = re.compile(r"^/(board)?game/(\d+).*$")
REGEX_BGG_USER = re.compile(r"^/user/([^/]+).*$")
REGEX_WIKIDATA_ID = re.compile(r"^/(wiki|entity|resource)/Q(\d+).*$")
REGEX_DBPEDIA_DOMAIN = re.compile(r"^[a-z]{2}\.dbpedia\.org$")
REGEX_DBPEDIA_ID = re.compile(r"^/(resource|page)/(.+)$")
REGEX_LUDING_ID = re.compile(r"^.*gameid/(\d+).*$")
REGEX_SPIELEN_ID = re.compile(
    r"^/(alle-brettspiele|messeneuheiten|ausgezeichnet-\d+)/(\w[^/]*).*$"
)
REGEX_FREEBASE_ID = re.compile(r"^/ns/(g|m)\.([^/]+).*$")
REGEX_BGA_ID = re.compile(r"^.*/game/([a-zA-Z0-9]+)(/.*)?$")


def to_lower(string):
    """safely convert to lower case string, else return None"""
    string = to_str(string)
    return string.lower() if string is not None else None


def identity(obj: Any) -> Any:
    """do nothing"""
    return obj


def _replace_utf_entities(match):
    try:
        values = tuple(map(parse_int, REGEX_SINGLE_ENT.findall(match.group(0))))
        bytes_ = bytes(values) if all(values) else None
        return bytes_.decode() if bytes_ else match.group(0)
    except Exception:
        pass
    return match.group(0)


def replace_utf_entities(string):
    """replace XML entities weirdly encoded"""
    return REGEX_ENTITIES.sub(_replace_utf_entities, string)


def replace_all_entities(string):
    """replace all XML entities, even poorly encoded"""
    # hack because BGG encodes 'Ü' as '&amp;#195;&amp;#156;' (d'oh!)
    # note that this may corrupt text that's actually encoded correctly!
    return replace_entities(
        replace_utf_entities(
            string.replace("&amp;", "&").replace("&amp;", "&").replace("&amp;", "&")
        )
    )


def extract_query_param(url: Union[str, ParseResult], field: str) -> Optional[str]:
    """extract a specific field from URL query parameters"""

    url = urlparse(url) if isinstance(url, str) else url
    query = parse_qs(url.query)
    values = query.get(field)

    return values[0] if values else None


def now(tzinfo=None):
    """current time in UTC or given timezone"""

    result = datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc)
    return result if tzinfo is None else result.astimezone(tzinfo)


def serialize_date(date: Any, tzinfo: Optional[timezone] = None) -> Optional[str]:
    """seralize a date into ISO format if possible"""

    parsed = parse_date(date, tzinfo)
    return parsed.strftime("%Y-%m-%dT%T%z") if parsed else str(date) if date else None


def parse_json(file_or_string, **kwargs):
    """safely parse JSON string"""

    if file_or_string is None:
        return None

    try:
        return json.load(file_or_string, **kwargs)
    except Exception:
        pass

    try:
        return json.loads(to_str(file_or_string), **kwargs)
    except Exception:
        pass

    return None


def _json_default(obj):
    if isinstance(obj, BaseItem):
        return dict(obj)
    if isinstance(obj, (set, frozenset, range, GeneratorType)) or hasattr(
        obj, "__iter__"
    ):
        return list(obj)
    if isinstance(obj, datetime):
        return serialize_date(obj)
    return repr(obj)


def serialize_json(obj, file=None, **kwargs):
    """
    safely serialze JSON, turning iterables into lists, dates into ISO strings,
    and everything else into their representation
    """

    kwargs.setdefault("default", _json_default)

    if isinstance(file, (str, bytes)):
        LOGGER.info("opening file <%s> and writing JSON content", file)

        path_dir = os.path.abspath(os.path.split(file)[0])
        os.makedirs(path_dir, exist_ok=True)

        with open(file, "w") as json_file:
            return json.dump(obj, json_file, **kwargs)

    if file is not None:
        LOGGER.debug("writing JSON content to opened file pointer <%s>", file)
        return json.dump(obj, file, **kwargs)

    return json.dumps(obj, **kwargs)


def date_from_file(
    path: Union[bytes, str, os.PathLike, None],
    tzinfo: Optional[timezone] = None,
    format_str: Optional[str] = None,
) -> Optional[datetime]:
    """Parse a date from a file."""

    if not path:
        return None

    path = Path(path).resolve()
    LOGGER.info("Reading date from path <%s>", path)

    try:
        with path.open() as file_obj:
            date = normalize_space(next(file_obj, None))
    except Exception:
        date = None

    return parse_date(date=date, tzinfo=tzinfo, format_str=format_str)


def validate_range(value, lower=None, upper=None):
    """validate that the given value is between lower and upper"""
    try:
        if (lower is None or lower <= value) and (upper is None or value <= upper):
            return value
    except TypeError:
        pass
    return None


def json_from_response(response):
    """Parse JSON from respose if possible."""
    result = parse_json(response.text) if hasattr(response, "text") else None
    return result or {}


def extract_meta(response=None):
    """Extract meta object from response if possible."""
    if hasattr(response, "meta") and response.meta:
        return response.meta
    if hasattr(response, "request") and hasattr(response.request, "meta"):
        return response.request.meta or {}
    return {}


def extract_item(item=None, response=None, item_cls=Item):
    """Extract item from response if possible."""
    if item:
        return item
    meta = extract_meta(response)
    return meta.get("item") or item_cls()


def extract_url(item=None, response=None, default=None):
    """Extract URL from response if possible."""
    if item and item.get("url"):
        return item["url"]
    meta = extract_meta(response)
    if meta.get("url"):
        return meta["url"]
    if hasattr(response, "url") and response.url:
        return response.url
    if hasattr(response, "request") and hasattr(response.request, "url"):
        return response.request.url
    return default


def _match(string: str, comparison: Union[str, Pattern]):
    return (
        string == comparison
        if isinstance(comparison, str)
        else bool(comparison.match(string))
    )


def parse_url(
    url: Union[str, ParseResult, None],
    hostnames: Optional[Iterable[Union[str, Pattern]]] = None,
) -> Optional[ParseResult]:
    """parse URL and optionally filter for hosts"""
    url = urlparse(url) if isinstance(url, str) else url
    hostnames = tuple(arg_to_iter(hostnames))
    return (
        url
        if url
        and url.hostname
        and url.path
        and (
            not hostnames
            or any(_match(url.hostname, hostname) for hostname in hostnames)
        )
        else None
    )


def validate_url(
    url: Union[str, ParseResult, None],
    hostnames: Optional[Iterable[Union[str, Pattern]]] = None,
    schemes: Optional[Iterable[Union[str, Pattern]]] = None,
) -> Optional[str]:
    """Returns cleaned up URL iff valid with scheme, hostname, and path."""
    url = parse_url(url=url, hostnames=hostnames)
    schemes = frozenset(arg_to_iter(schemes))
    return (
        url.geturl()
        if url is not None and url.scheme and (not schemes or url.scheme in schemes)
        else None
    )


def extract_bgg_id(url: Union[str, ParseResult, None]) -> Optional[int]:
    """extract BGG ID from URL"""
    url = parse_url(url, ("boardgamegeek.com", "www.boardgamegeek.com"))
    if not url:
        return None
    match = REGEX_BGG_ID.match(url.path)
    bgg_id = parse_int(match.group(2)) if match else None
    return bgg_id if bgg_id is not None else parse_int(extract_query_param(url, "id"))


def extract_bgg_user_name(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract BGG user name from url"""
    url = parse_url(url, ("boardgamegeek.com", "www.boardgamegeek.com"))
    if not url:
        return None
    match = REGEX_BGG_USER.match(url.path)
    user_name = (
        unquote_plus(match.group(1)) if match else extract_query_param(url, "username")
    )
    return user_name.lower() if user_name else None


def extract_wikidata_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract Wikidata ID from URL"""
    url = parse_url(url, ("wikidata.org", "www.wikidata.org", "wikidata.dbpedia.org"))
    if not url:
        return None
    match = REGEX_WIKIDATA_ID.match(url.path)
    return f"Q{match.group(2)}" if match else extract_query_param(url, "id")


def extract_wikipedia_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract Wikipedia ID from URL"""
    url = parse_url(url, ("en.wikipedia.org", "en.m.wikipedia.org"))
    return (
        unquote_plus(url.path[6:]) or None
        if url and url.path.startswith("/wiki/")
        else None
    )


def extract_dbpedia_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract DBpedia ID from URL"""
    url = parse_url(url, ("dbpedia.org", "www.dbpedia.org", REGEX_DBPEDIA_DOMAIN))
    if not url:
        return None
    match = REGEX_DBPEDIA_ID.match(url.path)
    return unquote_plus(match.group(2)) if match else extract_query_param(url, "id")


def extract_luding_id(url: Union[str, ParseResult, None]) -> Optional[int]:
    """extract Luding ID from URL"""
    url = parse_url(url, ("luding.org", "www.luding.org"))
    if not url:
        return None
    match = REGEX_LUDING_ID.match(url.path)
    return (
        parse_int(match.group(1))
        if match
        else parse_int(extract_query_param(url, "gameid"))
    )


def extract_spielen_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract Spielen.de ID from URL"""
    url = parse_url(
        url, ("gesellschaftsspiele.spielen.de", "www.gesellschaftsspiele.spielen.de")
    )
    if not url:
        return None
    match = REGEX_SPIELEN_ID.match(url.path)
    spielen_id = unquote_plus(match.group(2)) if match else None
    return (
        spielen_id if parse_int(spielen_id) is None else extract_query_param(url, "id")
    )


def extract_freebase_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract Freebase ID from URL"""
    url = parse_url(url, ("rdf.freebase.com", "freebase.com"))
    if not url:
        return None
    match = REGEX_FREEBASE_ID.match(url.path)
    return (
        f"/{match.group(1)}/{match.group(2)}"
        if match
        else extract_query_param(url, "id")
    )


def extract_bga_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """extract Board Game Atlas ID from URL"""
    url = parse_url(url, ("boardgameatlas.com", "www.boardgameatlas.com"))
    if not url:
        return None
    match = REGEX_BGA_ID.match(url.path)
    if match:
        return match.group(1)
    ids_str = extract_query_param(url, "ids")
    ids = ids_str.split(",") if ids_str else ()
    return take_first(map(normalize_space, ids)) or extract_query_param(url, "game-id")


def extract_ids(*urls: Optional[str]) -> Dict[str, List[Union[int, str]]]:
    """extract all possible IDs from all the URLs"""
    urls = tuple(map(urlparse, urls))
    return {
        "bgg_id": clear_list(map(extract_bgg_id, urls)),
        "freebase_id": clear_list(map(extract_freebase_id, urls)),
        "wikidata_id": clear_list(map(extract_wikidata_id, urls)),
        "wikipedia_id": clear_list(map(extract_wikipedia_id, urls)),
        "dbpedia_id": clear_list(map(extract_dbpedia_id, urls)),
        "luding_id": clear_list(map(extract_luding_id, urls)),
        "spielen_id": clear_list(map(extract_spielen_id, urls)),
        "bga_id": clear_list(map(extract_bga_id, urls)),
    }


@lru_cache(maxsize=8)
def pubsub_client():
    """Google Cloud PubSub client."""

    try:
        from google.cloud import pubsub

        return pubsub.SubscriberClient()

    except Exception:
        LOGGER.exception("unable to initialise PubSub client")

    return None



# -*- coding: utf-8 -*-

""" Luding spider """

import os
import re
import string

from pytility import parse_int
from scrapy import Spider
from scrapy.utils.misc import arg_to_iter

from ..items import GameItem
from ..loaders import GameLoader
from ..utils import extract_ids, extract_query_param


class LudingSpider(Spider):
    """Luding spider"""

    name = "luding"
    allowed_domains = ("luding.org",)
    start_urls = tuple(
        "http://luding.org/cgi-bin/GameFirstLetter.py?letter={}".format(letter)
        for letter in string.ascii_uppercase + "0"
    )
    item_classes = (GameItem,)

    custom_settings = {
        "DOWNLOAD_DELAY": 2,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 8,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 4,
        "LIMIT_IMAGES_TO_DOWNLOAD": parse_int(
            os.getenv("LIMIT_IMAGES_TO_DOWNLOAD_LUDING")
        )
        or 0,
    }

    def parse(self, response):
        """
        @url http://luding.org/cgi-bin/GameFirstLetter.py?letter=A
        @returns items 0 0
        @returns requests 2000
        """

        for game in response.css("table.game-list > tr"):
            url = game.xpath("td[1]//a/@href").extract_first()
            if url:
                yield response.follow(url, callback=self.parse_game)

    # pylint: disable=no-self-use
    def parse_game(self, response):
        """
        @url http://www.luding.org/cgi-bin/GameData.py/ENgameid/1508
        @returns items 1 1
        @returns requests 0 0
        @scrapes name year game_type description designer artist publisher \
                 url image_url review_url external_link luding_id \
                 min_players max_players min_age
        """

        headline = response.css("h1")
        game = headline.xpath("following-sibling::table")

        ldr = GameLoader(item=GameItem(), selector=game, response=response)

        ldr.add_value("name", headline.extract_first())
        ldr.add_xpath("year", 'tr[td = "Year:"]/td[2]')
        ldr.add_xpath("game_type", 'tr[td = "Type:"]/td[2]')
        ldr.add_xpath("description", 'tr[td = "Box text:"]/td[2]')

        ldr.add_xpath("designer", 'tr[td = "Designer:"]/td[2]/a')
        ldr.add_xpath("artist", 'tr[td = "Art:"]/td[2]/a')
        ldr.add_xpath("publisher", 'tr[td = "Publisher name:"]/td[2]/a')

        ldr.add_xpath("url", "(.//a/@href)[last()]")
        images = game.css("img::attr(src)").extract()
        ldr.add_value("image_url", {response.urljoin(i) for i in images})
        review_urls = game.xpath(
            'tr[contains(td[1], "review")]//'
            'a/@href[starts-with(., "/cgi-bin/Redirect.py")]'
        ).extract()
        review_urls = [
            extract_query_param(response.urljoin(link), "URL") for link in review_urls
        ]
        ldr.add_value("review_url", review_urls)
        links = game.xpath(
            './/a/@href[starts-with(., "/cgi-bin/Redirect.py")]'
        ).extract()
        links = (extract_query_param(response.urljoin(link), "URL") for link in links)
        links = [link for link in links if link not in frozenset(review_urls)]
        ldr.add_value("external_link", links)

        players = game.xpath('tr[td = "No. of players:"]/td[2]/text()').extract_first()
        players = players.split("-") if players else [None]
        ldr.add_value("min_players", players[0])
        ldr.add_value("max_players", players[-1])
        age = game.xpath('tr[td = "Age:"]/td[2]/text()').extract_first()
        age = re.match(r"^.*?(\d+).*$", age) if age else None
        ldr.add_value("min_age", age.group(1) if age else None)

        ldr.add_value(
            None,
            extract_ids(response.url, *arg_to_iter(links), *arg_to_iter(review_urls)),
        )

        return ldr.load_item()



# -*- coding: utf-8 -*-

"""BoardGameGeek hotness spider."""

from datetime import timezone
from pathlib import Path

from pytility import parse_date
from scrapy import Request, Spider

from ..items import GameItem
from ..loaders import GameLoader
from ..utils import now


class BggHotnessSpider(Spider):
    """BoardGameGeek hotness spider."""

    name = "bgg_hotness"
    allowed_domains = ("boardgamegeek.com",)
    start_urls = ("https://www.boardgamegeek.com/xmlapi2/hot?type=boardgame",)
    item_classes = (GameItem,)

    custom_settings = {
        "DOWNLOAD_DELAY": 0,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 1024,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 1024,
        "AUTOTHROTTLE_HTTP_CODES": (429, 503, 504),
    }

    def _local_requests(self, path_dir="."):
        path_dir = Path(path_dir).resolve()

        for path_file in path_dir.iterdir():
            if not path_file.is_file():
                continue

            self.logger.info("Processing <%s>", path_file)

            date = parse_date(path_file.stem, tzinfo=timezone.utc)

            yield Request(
                url=path_file.as_uri(), callback=self.parse, meta={"published_at": date}
            )

    def start_requests(self):
        """Initial requests, either locally or from BGG."""

        hotness_dir = self.settings.get("BGG_HOTNESS_DIR")

        if hotness_dir:
            self.logger.info("Loading local files from <%s>", hotness_dir)
            yield from self._local_requests(hotness_dir)

        else:
            yield from super().start_requests()

    def parse(self, response):
        """
        @url https://www.boardgamegeek.com/xmlapi2/hot?type=boardgame
        @returns items 50 50
        @returns requests 0 0
        @scrapes published_at rank bgg_id name year image_url scraped_at
        """

        scraped_at = now()

        for game in response.xpath("/items/item"):
            ldr = GameLoader(item=GameItem(), selector=game, response=response)

            ldr.add_xpath("bgg_id", "@id")
            ldr.add_xpath("rank", "@rank")
            ldr.add_xpath("name", "name/@value")
            ldr.add_xpath("year", "yearpublished/@value")
            ldr.add_xpath("image_url", "thumbnail/@value")

            ldr.add_value("published_at", response.meta.get("published_at"))
            ldr.add_value("published_at", scraped_at)
            ldr.add_value("scraped_at", scraped_at)

            yield ldr.load_item()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" make prefixes """

import argparse
import logging
import os
import re
import string
import sys

from itertools import groupby

from pytility import parse_int, to_str
from pytrie import SortedStringTrie as Trie
from scrapy.utils.misc import arg_to_iter

from .utils import parse_json, serialize_json

LOGGER = logging.getLogger(__name__)
NON_WORD_REGEX = re.compile(r"[^a-z]")
LIMIT = 1_000_000
LETTERS = string.ascii_lowercase + "_"


def _parse_key(key):
    key = key.lower()
    key, _ = NON_WORD_REGEX.subn("_", key)
    return key


def _process_file(file, fields="bgg_user_name", sep=",", count=False):
    fields = tuple(arg_to_iter(fields))

    if isinstance(file, (str, bytes, os.PathLike)):
        with open(file, "r") as file_obj:
            yield from _process_file(file_obj, fields, sep, count)
            return

    items = filter(None, map(parse_json, file))
    for key, group in groupby(
        items, key=lambda item: tuple(_parse_key(item[field]) for field in fields)
    ):
        yield sep.join(key), sum(1 for _ in group) if count else group

    try:
        file.seek(0)
    except Exception:
        LOGGER.error("unable to reset file position in <%s>...", file)


def _make_trie(file, fields="bgg_user_name", sep=","):
    return Trie(_process_file(file, fields, sep, count=True))


def _count(trie, prefix=None):
    return sum(trie.itervalues(prefix))


def _prefixes(trie, prefix="", limit=LIMIT):
    for letter in LETTERS:
        pre = prefix + letter
        count = _count(trie, pre)
        if count <= limit:
            yield pre, count
        else:
            yield from _prefixes(trie, pre, limit)


def _prefixes_from_file(file):
    if isinstance(file, (str, bytes, os.PathLike)):
        LOGGER.info("reading prefixes from <%s>...", file)
        with open(file) as file_obj:
            yield from _prefixes_from_file(file_obj)
            return

    for line in file:
        line = to_str(line)
        parts = line.split() if line else None
        if not parts or not parts[0]:
            continue
        count = parse_int(parts[1]) if len(parts) >= 2 else None
        yield parts[0], count


def _trie_from_file(file):
    try:
        return Trie(_prefixes_from_file(file))
    except Exception:
        pass
    return None


def _save_to_prefixes(dst, trie, file, fields="bgg_user_name", sep=","):
    seen = set()
    total = 0

    for prefix, group in groupby(
        _process_file(file, fields, sep, False),
        key=lambda item: trie.longest_prefix(item[0]),
    ):
        path = os.path.abspath(dst.format(prefix=prefix))
        os.makedirs(os.path.dirname(path), exist_ok=True)

        if prefix in seen:
            LOGGER.info("appending to file <%s>...", path)
            mode = "a"
        else:
            LOGGER.info("writing to file <%s>...", path)
            mode = "w"
            seen.add(prefix)

        count = 0
        with open(path, mode) as file_obj:
            for _, items in group:
                for item in items:
                    serialize_json(item, file_obj, indent=None, sort_keys=True)
                    file_obj.write("\n")
                    count += 1
        LOGGER.info("done writing %d items to <%s>...", count, path)
        total += count

    LOGGER.info("done writing %d items in total...", total)


def split_file(
    in_file,
    out_file,
    fields="bgg_user_name",
    trie_file=None,
    limits=LIMIT,
    construct=False,
):
    """split input file along prefixes"""

    trie = None

    if trie_file and not construct:
        LOGGER.info("loading trie from file <%s>...", trie_file)
        trie = _trie_from_file(trie_file)

    if not trie:
        LOGGER.info("making trie for <%s>...", in_file)
        full_trie = _make_trie(file=in_file, fields=fields)
        limits = tuple(arg_to_iter(limits)) or (LIMIT,)

        for limit in limits:
            trie = Trie(_prefixes(full_trie, limit=limit))
            LOGGER.info("%d prefixes using limit %d", len(trie), limit)
            out_path = trie_file.format(limit=limit) if trie_file else None
            if not out_path or out_path == "-":
                for prefix, count in trie.items():
                    print(f"{prefix}\t{count}")
            else:
                with open(out_path, "w") as file_obj:
                    for prefix, count in trie.items():
                        file_obj.write(f"{prefix}\t{count}\n")

    LOGGER.info("constructed trie of size %d", len(trie))
    _save_to_prefixes(dst=out_file, trie=trie, file=in_file, fields=fields)


def _parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("in_path", help="input JSON Lines file")
    parser.add_argument(
        "--construct", "-c", action="store_true", help="construct a new trie"
    )
    parser.add_argument("--trie-path", "-t", help="path to trie")
    parser.add_argument("--limits", "-l", nargs="+", type=int, help="limits to use")
    parser.add_argument(
        "--keys",
        "-k",
        nargs="+",
        default=("bgg_user_name",),
        help="target columns for merging",
    )
    parser.add_argument("--out-path", "-o", help="output path")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def _main():
    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    split_file(
        in_file=args.in_path,
        out_file=args.out_path,
        fields=args.keys,
        trie_file=args.trie_path,
        limits=args.limits,
        construct=args.construct,
    )


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Split JSONL files."""

import argparse
import json
import logging
import os
import sys

from pathlib import Path

from pytility import batchify
from scrapy.utils.misc import arg_to_iter

try:
    # pylint: disable=redefined-builtin
    from smart_open import open
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)
FIELDS = frozenset(
    {
        "article_id",
        "url_canonical",
        "url_mobile",
        "url_amp",
        "url_thumbnail",
        "published_at",
        "title_short",
        "author",
        "description",
        "category",
        "keyword",
        "section_inferred",
        "country",
        "language",
        "source_name",
    }
)


def _is_empty(item):
    return (
        isinstance(item, (bytes, dict, frozenset, list, set, str, tuple)) and not item
    )


def _filter_fields(item, fields=None, exclude_empty=False):
    return {
        k: v
        for k, v in item.items()
        if (not fields or k in fields) and (not exclude_empty or not _is_empty(v))
    }


def _load_items(iterable, fields=None, exclude_empty=False):
    if isinstance(iterable, (bytes, str, os.PathLike)):
        LOGGER.info("Reading from <%s>", iterable)
        with open(iterable) as file_obj:
            yield from _load_items(file_obj, fields=fields, exclude_empty=exclude_empty)
        return

    fields = frozenset(arg_to_iter(fields))

    for i, line in enumerate(iterable):
        try:
            item = json.loads(line)
        except json.JSONDecodeError:
            LOGGER.exception("Unable to parse line %d: %s [...]", i + 1, line[:100])
        else:
            yield _filter_fields(item=item, fields=fields, exclude_empty=exclude_empty)


def split_files(
    path_in,
    path_out=None,
    size=None,
    fields=FIELDS,
    exclude_empty=False,
    dry_run: bool = False,
):
    """Split a JSON lines file into JSON files of a given size."""

    dry_run_prefix = "[DRY RUN] " if dry_run else ""

    path_in = Path(path_in).resolve()
    path_out = "-" if path_out is None or path_out == "-" else Path(path_out).resolve()

    LOGGER.info(
        "%sReading items from <%s> splitting them into <%s>",
        dry_run_prefix,
        path_in,
        path_out,
    )

    if not dry_run and path_out != "-":
        path_out.parent.mkdir(parents=True, exist_ok=True)

    items = tuple(_load_items(path_in, fields=fields, exclude_empty=exclude_empty))
    batches = batchify(items, size) if size else (items,)
    total = len(items)
    count = 0

    LOGGER.info("%sRead %d items from <%s>", dry_run_prefix, total, path_in)

    for i, batch in enumerate(batches):
        batch = list(batch)
        count += len(batch)
        result = {
            "count": total,
            "previous": i - 1 if i else None,
            "next": i + 1 if count < total else None,
            "results": batch,
        }

        if path_out == "-":
            json.dump(result, sys.stdout, sort_keys=True)
            print()

        else:
            out_path = str(path_out).format(number=i)
            LOGGER.info("%sWriting batch #%d to <%s>", dry_run_prefix, i, out_path)
            if not dry_run:
                with open(out_path, "w") as out_file:
                    json.dump(result, out_file, sort_keys=True)

    LOGGER.info("%sDone splitting.", dry_run_prefix)


def _parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("infile", help="input file")
    parser.add_argument("--batch", "-b", type=int, help="batch size")
    parser.add_argument("--outfile", "-o", help="output file path")
    parser.add_argument("--dry-run", "-n", action="store_true", help="dry run")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def _main():
    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    split_files(
        path_in=args.infile,
        path_out=args.outfile,
        size=args.batch,
        fields=FIELDS,
        exclude_empty=True,
        dry_run=args.dry_run,
    )


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Pull logs from Google Cloud PubSub queue."""

import argparse
import csv
import logging
import os
import sys

from functools import partial
from itertools import count
from time import sleep
from typing import Optional

from pytility import normalize_space

try:
    # pylint: disable=redefined-builtin
    from smart_open import open
except ImportError:
    pass

from .utils import now, pubsub_client

LOGGER = logging.getLogger(__name__)


def _process_messages_csv(
    *,
    messages,
    output,
    header=False,
    message_col="message",
    message_process=None,
    encoding="utf-8",
):
    writer = csv.writer(output)

    if header:
        writer.writerow(("date", message_col))

    for message in messages:
        try:
            date = message.message.publish_time.replace(nanosecond=0).isoformat()
            content = message.message.data.decode(encoding)
            if callable(message_process):
                content = message_process(content)
            if date and content:
                writer.writerow((date, content))
                yield message.ack_id
            else:
                LOGGER.error("there was a problem processing message %r", message)

        except Exception:
            LOGGER.exception("unable to process message %r", message)


def _process_messages_raw(
    *,
    messages,
    output,
    header=False,
    message_process=None,
    encoding="utf-8",
):
    del header  # required to align with _process_messages_csv
    for message in messages:
        try:
            content = message.message.data.decode(encoding)
            if callable(message_process):
                content = message_process(content)
            if content:
                output.write(content)
                output.write("\n")
                yield message.ack_id
            else:
                LOGGER.error("there was a problem processing message %r", message)

        except Exception:
            LOGGER.exception("unable to process message %r", message)


def _format_from_path(path: Optional[str]) -> Optional[str]:
    if not path:
        return None
    ext = path.rsplit(".", 1)[-1]
    return ext.lower()


def _parse_args():
    parser = argparse.ArgumentParser(
        description="Pull logs from Google Cloud PubSub queue."
    )
    parser.add_argument(
        "--project",
        "-p",
        default=os.getenv("PULL_QUEUE_PROJECT"),
        help="Google Cloud project",
    )
    parser.add_argument(
        "--subscription",
        "-s",
        default=os.getenv("PULL_QUEUE_SUBSCRIPTION_LOGS"),
        help="Google Cloud PubSub subscription",
    )
    parser.add_argument(
        "--out-path",
        "-o",
        help="Output location",
    )
    parser.add_argument(
        "--format",
        "-f",
        choices=("csv", "raw"),
        help="Output format",
    )
    parser.add_argument(
        "--header",
        "-H",
        action="store_true",
        help="include CSV header",
    )
    parser.add_argument(
        "--message-col",
        "-C",
        help="Label of the message column",
    )
    parser.add_argument(
        "--message-process",
        "-P",
        choices=("lower", "normalize"),
        help="How to process the message",
    )
    parser.add_argument(
        "--batch-size",
        "-b",
        type=int,
        default=1000,
        help="maximum batch size (max messages per pull and file)",
    )
    parser.add_argument(
        "--timeout",
        "-t",
        type=float,
        default=60,
        help="timeout for a pull in seconds",
    )
    parser.add_argument(
        "--sleep",
        "-S",
        type=float,
        help="sleep for that many seconds before start pulling",
    )
    parser.add_argument(
        "--no-ack",
        "-n",
        action="store_true",
        help="do not acknowledge messages to subscription",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Pull logs from Google Cloud PubSub queue."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    if not args.project or not args.subscription:
        LOGGER.error("Google Cloud PubSub project and subscription are required")
        sys.exit(1)

    output_format = args.format or _format_from_path(args.out_path) or "raw"
    LOGGER.info("Using output format <%s>", output_format)

    message_col = args.message_col or "message"

    if args.message_process == "lower":
        message_process = lambda m: normalize_space(m).lower()
    elif args.message_process == "normalize":
        message_process = normalize_space
    else:
        message_process = None

    process_messages = (
        partial(
            _process_messages_csv,
            message_col=message_col,
            message_process=message_process,
        )
        if output_format == "csv"
        else partial(_process_messages_raw, message_process=message_process)
    )

    if args.sleep:
        LOGGER.info("going to sleep for %.1f seconds", args.sleep)
        sleep(args.sleep)

    client = pubsub_client()
    # pylint: disable=no-member
    subscription_path = client.subscription_path(args.project, args.subscription)

    for i in count():
        LOGGER.info("#%d: pulling subscription <%s>", i, subscription_path)

        try:
            response = client.pull(
                subscription=subscription_path,
                max_messages=args.batch_size,
                return_immediately=False,
                timeout=args.timeout,
            )
        except Exception:
            LOGGER.info("subscription <%s> timed out", subscription_path)
            break

        if not response or not response.received_messages:
            LOGGER.info(
                "nothing to be pulled from subscription <%s>", subscription_path
            )
            break

        if not args.out_path or args.out_path == "-":
            ack_ids = tuple(
                process_messages(
                    messages=response.received_messages,
                    output=sys.stdout,
                    header=args.header and (i == 0),
                )
            )
        else:
            curr = now()
            out_path = args.out_path.format(
                number=i,
                datetime=curr.strftime("%Y-%m-%dT%H-%M-%S"),
                date=curr.strftime("%Y-%m-%d"),
                time=curr.strftime("%H-%M-%S"),
                year=curr.year,
                month=curr.month,
                day=curr.day,
                hour=curr.hour,
                minute=curr.minute,
                second=curr.second,
            )
            LOGGER.info("writing results to <%s>", out_path)
            with open(out_path, "w", newline="") as out_file:
                ack_ids = tuple(
                    process_messages(
                        messages=response.received_messages,
                        output=out_file,
                        header=args.header,
                    )
                )

        LOGGER.info("%d message(s) successfully processed", len(ack_ids))

        if ack_ids and not args.no_ack:
            LOGGER.info(
                "acknowledge %d message(s) in subscription <%s>",
                len(ack_ids),
                subscription_path,
            )
            client.acknowledge(subscription=subscription_path, ack_ids=ack_ids)

    LOGGER.info("Done.")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

""" Scrapy item loaders """

from pytility import normalize_space
from scrapy.loader import ItemLoader
from scrapy.loader.processors import TakeFirst, MapCompose
from scrapy_extensions import JsonLoader
from w3lib.html import remove_tags

from .utils import identity, replace_all_entities


class GameLoader(ItemLoader):
    """loader for GameItem"""

    default_input_processor = MapCompose(
        identity, str, remove_tags, replace_all_entities, normalize_space
    )
    default_output_processor = TakeFirst()


class GameJsonLoader(JsonLoader, GameLoader):
    """loader for GameItem plus JMESPath capabilities"""


class UserLoader(ItemLoader):
    """loader for UserItem"""

    default_input_processor = MapCompose(
        identity, str, remove_tags, replace_all_entities, normalize_space
    )
    default_output_processor = TakeFirst()


class RatingLoader(ItemLoader):
    """loader for RatingItem"""

    default_input_processor = MapCompose(
        identity, str, remove_tags, replace_all_entities, normalize_space
    )
    default_output_processor = TakeFirst()


class RatingJsonLoader(JsonLoader, RatingLoader):
    """loader for RatingItem plus JMESPath capabilities"""



# -*- coding: utf-8 -*-

""" Scrapy settings """

import os

from pytility import parse_bool

try:
    from dotenv import find_dotenv, load_dotenv

    DOTENV_PATH = os.getenv("DOTENV_PATH") or find_dotenv(usecwd=True)
    load_dotenv(dotenv_path=DOTENV_PATH, verbose=True)

except ImportError:
    pass

BOT_NAME = "board-games"

SPIDER_MODULES = ["board_game_scraper.spiders"]
NEWSPIDER_MODULE = "board_game_scraper.spiders"

LOG_LEVEL = os.getenv("LOG_LEVEL") or "INFO"
LOG_FORMATTER = "scrapy_extensions.QuietLogFormatter"
LOG_SCRAPED_ITEMS = parse_bool(os.getenv("LOG_SCRAPED_ITEMS"))

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

FEED_EXPORT_FIELDS = (
    "name",
    "alt_name",
    "year",
    "game_type",
    "description",
    "designer",
    "artist",
    "publisher",
    "url",
    "official_url",
    "image_url",
    "image_file",
    "video_url",
    "rules_url",
    "rules_file",
    "review_url",
    "external_link",
    "list_price",
    "min_players",
    "max_players",
    "min_players_rec",
    "max_players_rec",
    "min_players_best",
    "max_players_best",
    "min_age",
    "max_age",
    "min_age_rec",
    "max_age_rec",
    "min_time",
    "max_time",
    "category",
    "mechanic",
    "cooperative",
    "compilation",
    "compilation_of",
    "family",
    "expansion",
    "implementation",
    "integration",
    "rank",
    "add_rank",
    "num_votes",
    "avg_rating",
    "stddev_rating",
    "bayes_rating",
    "worst_rating",
    "best_rating",
    "complexity",
    "easiest_complexity",
    "hardest_complexity",
    "language_dependency",
    "lowest_language_dependency",
    "highest_language_dependency",
    "bgg_id",
    "freebase_id",
    "wikidata_id",
    "wikipedia_id",
    "dbpedia_id",
    "luding_id",
    "spielen_id",
    "bga_id",
    "published_at",
    "updated_at",
    "scraped_at",
)

MULTI_FEED_ENABLED = True
MULTI_FEED_EXPORT_FIELDS = {
    "GameItem": FEED_EXPORT_FIELDS,
    "UserItem": (
        "item_id",
        "bgg_user_name",
        "first_name",
        "last_name",
        "registered",
        "last_login",
        "country",
        "region",
        "city",
        "external_link",
        "image_url",
        "image_file",
        "published_at",
        "updated_at",
        "scraped_at",
    ),
    "RatingItem": (
        "item_id",
        "bgg_id",
        "bgg_user_name",
        "bgg_user_rating",
        "bgg_user_owned",
        "bgg_user_prev_owned",
        "bgg_user_for_trade",
        "bgg_user_want_in_trade",
        "bgg_user_want_to_play",
        "bgg_user_want_to_buy",
        "bgg_user_preordered",
        "bgg_user_wishlist",
        "bgg_user_play_count",
        "bga_id",
        "bga_user_id",
        "bga_user_name",
        "bga_user_rating",
        "comment",
        "published_at",
        "updated_at",
        "scraped_at",
    ),
}

# Crawl responsibly by identifying yourself (and your website) on the user-agent
# USER_AGENT = 'board-games (+http://www.yourdomain.com)'
# USER_AGENT = ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) '
#               'AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36')

# Obey robots.txt rules
ROBOTSTXT_OBEY = parse_bool(os.getenv("ROBOTSTXT_OBEY") or True)
ROBOTSTXT_PARSER = "scrapy.robotstxt.PythonRobotParser"

# Configure maximum concurrent requests performed by Scrapy (default: 16)
# CONCURRENT_REQUESTS = 32

# Configure a delay for requests for the same website (default: 0)
# See http://scrapy.readthedocs.org/en/latest/topics/settings.html#download-delay
# See also autothrottle settings and docs
DOWNLOAD_DELAY = 0.1
# The download delay setting will honor only one of:
CONCURRENT_REQUESTS_PER_DOMAIN = 8
# CONCURRENT_REQUESTS_PER_IP = 16

# Disable cookies (enabled by default)
# COOKIES_ENABLED = False

# Disable Telnet Console (enabled by default)
# TELNETCONSOLE_ENABLED = False

# Override the default request headers:
DEFAULT_REQUEST_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Language": "en",
}

# Enable or disable spider middlewares
# See http://scrapy.readthedocs.org/en/latest/topics/spider-middleware.html
# SPIDER_MIDDLEWARES = {
#    'board_game_scraper.middlewares.MyCustomSpiderMiddleware': 543,
# }

# Enable or disable downloader middlewares
# See http://scrapy.readthedocs.org/en/latest/topics/downloader-middleware.html
DOWNLOADER_MIDDLEWARES = {"scrapy_extensions.DelayedRetry": 555}

# Enable or disable extensions
# See http://scrapy.readthedocs.org/en/latest/topics/extensions.html
EXTENSIONS = {
    "scrapy.extensions.closespider.CloseSpider": 0,
    "scrapy.extensions.feedexport.FeedExporter": None,
    "scrapy_extensions.MultiFeedExporter": 0,
    "scrapy.extensions.throttle.AutoThrottle": None,
    "scrapy_extensions.NicerAutoThrottle": 0,
    "board_game_scraper.extensions.StateTag": 0,
    "board_game_scraper.extensions.DontRunBeforeTag": 0,
    "board_game_scraper.extensions.PullQueueExtension": 100,
    "scrapy_extensions.MonitorDownloadsExtension": 500,
    "scrapy_extensions.DumpStatsExtension": 500,
}

# Configure item pipelines
# See http://scrapy.readthedocs.org/en/latest/topics/item-pipeline.html
ITEM_PIPELINES = {
    "board_game_scraper.pipelines.DataTypePipeline": 100,
    "scrapy_extensions.ValidatePipeline": 200,
    "board_game_scraper.pipelines.ResolveLabelPipeline": 300,
    "board_game_scraper.pipelines.ResolveImagePipeline": 400,
    "board_game_scraper.pipelines.LimitImagesPipeline": 500,
    "scrapy.pipelines.images.ImagesPipeline": 600,
    "scrapy.pipelines.images.FilesPipeline": None,
    "board_game_scraper.pipelines.CleanItemPipeline": 900,
}

# See https://doc.scrapy.org/en/latest/topics/extensions.html#module-scrapy.extensions.closespider
CLOSESPIDER_TIMEOUT = os.getenv("CLOSESPIDER_TIMEOUT")

# Enable and configure the AutoThrottle extension (disabled by default)
# See http://doc.scrapy.org/en/latest/topics/autothrottle.html
AUTOTHROTTLE_ENABLED = True
# The initial download delay
AUTOTHROTTLE_START_DELAY = max(DOWNLOAD_DELAY * 2, 5)
# The maximum download delay to be set in case of high latencies
AUTOTHROTTLE_MAX_DELAY = 60
# The average number of requests Scrapy should be sending in parallel to
# each remote server
AUTOTHROTTLE_TARGET_CONCURRENCY = CONCURRENT_REQUESTS_PER_DOMAIN
# Enable showing throttling stats for every response received:
AUTOTHROTTLE_DEBUG = False
AUTOTHROTTLE_HTTP_CODES = (429,)

# Enable and configure HTTP caching (disabled by default)
# See http://scrapy.readthedocs.org/en/latest/topics/downloader-middleware.html
HTTPCACHE_ENABLED = True
HTTPCACHE_EXPIRATION_SECS = 60 * 60 * 24 * 7  # 1 week
# HTTPCACHE_DIR = 'httpcache'
HTTPCACHE_IGNORE_HTTP_CODES = (500, 502, 503, 504, 408, 429, 202)
# HTTPCACHE_STORAGE = 'scrapy.extensions.httpcache.FilesystemCacheStorage'
HTTPCACHE_POLICY = "scrapy.extensions.httpcache.RFC2616Policy"

RETRY_ENABLED = True
RETRY_HTTP_CODES = (500, 502, 503, 504, 408, 429)

DELAYED_RETRY_ENABLED = False
DELAYED_RETRY_TIMES = -1
DELAYED_RETRY_HTTP_CODES = ()
DELAYED_RETRY_DELAY = 10.0
DELAYED_RETRY_PRIORITY_ADJUST = 0
DELAYED_RETRY_BACKOFF = True
DELAYED_RETRY_BACKOFF_MAX_DELAY = 100.0

# Monitoring settings
MONITOR_DOWNLOADS_ENABLED = True
MONITOR_DOWNLOADS_INTERVAL = 60
DUMP_STATS_ENABLED = True
DUMP_STATS_INTERVAL = 5 * 60

# Spider settings
SCRAPE_BGG_RATINGS = True
SCRAPE_BGG_COLLECTIONS = True
SCRAPE_BGG_USERS = True

# State tags
STATE_TAG_FILE = ".state"
PID_TAG_FILE = ".pid"

# "Don't run before" settings
DONT_RUN_BEFORE_FILE = os.getenv("DONT_RUN_BEFORE_FILE")
DONT_RUN_BEFORE_SEC = os.getenv("DONT_RUN_BEFORE_SEC")
DONT_RUN_BEFORE_DATE = os.getenv("DONT_RUN_BEFORE_DATE")

MEDIA_ALLOW_REDIRECTS = True

# LimitImagesPipeline
LIMIT_IMAGES_TO_DOWNLOAD = 0
LIMIT_IMAGES_URLS_FIELD = "image_url"

# Image processing
IMAGES_STORE = os.path.join(BASE_DIR, "images")
IMAGES_URLS_FIELD = "image_url_download"
IMAGES_RESULT_FIELD = "image_file"
IMAGES_EXPIRES = 360
# IMAGES_THUMBS = {"thumb": (1024, 1024)}

# File processing
FILES_STORE = os.path.join(BASE_DIR, "rules")
FILES_URLS_FIELD = "rules_url"
FILES_RESULT_FIELD = "rules_file"
FILES_EXPIRES = 180

# CleanItemPipeline
CLEAN_ITEM_DROP_FALSEY = True
CLEAN_ITEM_DROP_VALUES = None

# Board Game Atlas
BGA_CLIENT_ID = os.getenv("BGA_CLIENT_ID")
BGA_SCRAPE_IMAGES = False
BGA_SCRAPE_VIDEOS = False
BGA_SCRAPE_REVIEWS = False

PULL_QUEUE_ENABLED = False
PULL_QUEUE_PROJECT = os.getenv("PULL_QUEUE_PROJECT")
PULL_QUEUE_SUBSCRIPTION = os.getenv("PULL_QUEUE_SUBSCRIPTION")
PULL_QUEUE_INTERVAL = os.getenv("PULL_QUEUE_INTERVAL") or 5 * 60  # 5 minutes
PULL_QUEUE_MAX_MESSAGES = os.getenv("PULL_QUEUE_MAX_MESSAGES") or 100
PULL_QUEUE_PREVENT_RESCRAPE_FOR = (
    os.getenv("PULL_QUEUE_PREVENT_RESCRAPE_FOR") or 6 * 60 * 60  # 6 hours
)
PULL_QUEUE_PULL_TIMEOUT = os.getenv("PULL_QUEUE_PULL_TIMEOUT") or 5  # 5 seconds



# -*- coding: utf-8 -*-

""" monkey-patches """

import csv
import sys

from scrapy.http import XmlResponse
from scrapy.responsetypes import responsetypes

from .__version__ import VERSION, __version__

csv.field_size_limit(sys.maxsize)

# monkey-patching responsetypes to include SPAQRL XML results
responsetypes.classes["application/sparql-results+xml"] = XmlResponse



# -*- coding: utf-8 -*-

"""Zip ranking files."""

import argparse
import logging
import sys
import zipfile

from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Optional, Union

from pytility import parse_date

from .utils import now

BASE_DIR = Path(__file__).resolve().parent.parent
DUMMY_DATE = datetime(1970, 1, 1)
DATE_FORMAT = "%Y-%m-%dT%H-%M-%S"
LOGGER = logging.getLogger(__name__)


def zip_ranking_files(
    *,
    rankings_dir: Union[Path, str],
    rankings_file_glob: str,
    output_file: Union[Path, str],
    dry_run: bool = False,
) -> None:
    """Zip the ranking files in a given directory."""

    rankings_dir = Path(rankings_dir).resolve()
    output_file = Path(output_file).resolve()

    LOGGER.info(
        "Compressing rankings file in <%s> matching glob <%s>, storing the output to <%s>",
        rankings_dir,
        rankings_file_glob,
        output_file,
    )

    if dry_run:
        for rankings_file in rankings_dir.glob(rankings_file_glob):
            LOGGER.info("[Dry run] Compressing file <%s>…", rankings_file)
        LOGGER.info("Done.")
        return

    with zipfile.ZipFile(
        file=output_file,
        mode="x",
        compression=zipfile.ZIP_DEFLATED,
    ) as zip_file:
        for rankings_file in rankings_dir.glob(rankings_file_glob):
            LOGGER.info("Compressing file <%s>…", rankings_file)
            zip_file.write(
                filename=rankings_file,
                arcname=rankings_file.relative_to(rankings_dir),
            )

    LOGGER.info("Done.")


def file_date(
    path: Union[Path, str],
    *,
    format_str: Optional[str] = None,
) -> Optional[datetime]:
    """Get the date from a given file, either from its name or the access stats."""

    path = Path(path)

    if format_str:
        # TODO not the most efficient way to do this
        dummy_date = DUMMY_DATE.strftime(format_str)

        date_from_name = parse_date(
            date=path.name[: len(dummy_date)],
            tzinfo=timezone.utc,
            format_str=format_str,
        )

        if date_from_name is not None:
            return date_from_name

    file_stats = path.stat()
    return (
        parse_date(date=file_stats.st_ctime, tzinfo=timezone.utc)
        or parse_date(date=file_stats.st_mtime, tzinfo=timezone.utc)
        or parse_date(date=file_stats.st_atime, tzinfo=timezone.utc)
    )


def delete_older_files(
    *,
    dir_path: Union[Path, str],
    file_glob: str,
    older_than: timedelta,
    dry_run: bool = False,
):
    """Delete files in a directory older than a given age."""
    dir_path = Path(dir_path).resolve()
    cutoff_date = now() - older_than

    LOGGER.info(
        "Deleting file in <%s> matching glob <%s> older than <%s>",
        dir_path,
        file_glob,
        cutoff_date,
    )

    for file_path in dir_path.glob(file_glob):
        if file_date(path=file_path, format_str=DATE_FORMAT) < cutoff_date:
            if dry_run:
                LOGGER.info("[Dry run] Deleting <%s>…", file_path)
            else:
                LOGGER.info("Deleting <%s>…", file_path)
                file_path.unlink()

    LOGGER.info("Done.")


def _parse_args():
    parser = argparse.ArgumentParser(description="Zip ranking files.")
    parser.add_argument(
        "--out-dir",
        "-d",
        default=BASE_DIR / "backup",
        help="Output directory",
    )
    parser.add_argument(
        "--out-file",
        "-f",
        default="bgg-rankings-%Y-%m-%d.zip",
        help="Output file name",
    )
    parser.add_argument(
        "--in-dir",
        "-i",
        default=BASE_DIR / "feeds",
        help="Input directory",
    )
    parser.add_argument(
        "--glob",
        "-g",
        default="bgg_rankings*/GameItem/*.jl",
        help="File glob for input files",
    )
    parser.add_argument(
        "--delete-files-older-than-days",
        "-D",
        type=int,
        help="If given, files older than this will be deleted after zipping",
    )
    parser.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="Dry run, don't write or delete anything",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    out_file = now().strftime(args.out_file)
    out_dir = Path(args.out_dir).resolve()
    out_path = out_dir / out_file

    if not args.dry_run:
        if out_path.exists():
            LOGGER.info("Output file <%s> already exists, aborting…", out_path)
            sys.exit(1)
        out_dir.mkdir(parents=True, exist_ok=True)

    zip_ranking_files(
        rankings_dir=args.in_dir,
        rankings_file_glob=args.glob,
        output_file=out_path,
        dry_run=args.dry_run,
    )

    if args.delete_files_older_than_days:
        delete_older_files(
            dir_path=args.in_dir,
            file_glob=args.glob,
            older_than=timedelta(days=args.delete_files_older_than_days),
            dry_run=args.dry_run,
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Read library __version__ and write to .env."""

import argparse
import logging
import sys

from pathlib import Path
from shutil import copyfileobj
from tempfile import TemporaryFile

from .__version__ import __version__

LOGGER = logging.getLogger(__name__)


def _parse_args():
    parser = argparse.ArgumentParser(
        description="Read library __version__ and write to .env."
    )
    parser.add_argument(
        "--target",
        "-t",
        default=".env",
        help="target file",
    )
    parser.add_argument(
        "--variable",
        "-V",
        default="LIBRARY_VERSION",
        help="variable name",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Read library __version__ and write to .env."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    target = Path(args.target).resolve()

    LOGGER.info("Trying to write version %s to file <%s>...", __version__, target)

    if not target.exists():
        LOGGER.error("Target file <%s> does not exist, aborting", target)
        sys.exit(1)

    LOGGER.info("Looking for environment variable <%s>", args.variable)

    with TemporaryFile("w+") as temp:
        with target.open() as in_file:
            for line in in_file:
                if line.startswith(f"{args.variable}="):
                    temp.write(f"{args.variable}={__version__}\n")
                else:
                    temp.write(line)

        temp.seek(0)

        with target.open("w") as out_file:
            copyfileobj(temp, out_file)

    LOGGER.info("Done.")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Stop scraper, merge files, delete old files, and resume scraping."""

import argparse
import logging
import subprocess
import sys

from datetime import timedelta
from pathlib import Path
from time import sleep

from pytility import parse_bool, parse_float
from yaml import safe_load

from .merge import merge_files
from .utils import now

LOGGER = logging.getLogger(__name__)
BASE_DIR = Path(__file__).resolve().parent.parent
FEEDS_DIR = BASE_DIR / "feeds"
DATA_DIR = (BASE_DIR / ".." / "board-game-data").resolve()


def merge_config(
    spider, item="GameItem", in_paths=None, out_path=None, full=False, **kwargs
):
    """Returns arguments for merging a given spider."""

    curr_date = now()
    curr_date_str = curr_date.strftime("%Y-%m-%dT%H-%M-%S")

    kwargs["in_paths"] = in_paths or FEEDS_DIR / spider / item / "*"
    kwargs.setdefault("keys", f"{spider}_id")
    kwargs.setdefault("key_types", "int" if spider in ("bgg", "luding") else "str")
    kwargs.setdefault("latest", "scraped_at")
    kwargs.setdefault("latest_types", "date")
    kwargs.setdefault("latest_min", curr_date - timedelta(days=90))
    kwargs.setdefault("concat_output", True)

    if parse_bool(full):
        kwargs["out_path"] = (
            out_path or FEEDS_DIR / spider / item / f"{curr_date_str}-merged.jl"
        )

    else:
        kwargs["out_path"] = out_path or DATA_DIR / "scraped" / f"{spider}_{item}.jl"
        kwargs.setdefault(
            "fieldnames_exclude",
            ("published_at", "updated_at", "scraped_at"),
        )
        kwargs.setdefault("sort_keys", True)

    return kwargs


def merge_configs(spider, full=False):
    """Yields configs for all items in a given spider."""

    full = parse_bool(full)

    if spider == "bga":
        yield merge_config(spider="bga", item="GameItem", full=full)
        yield merge_config(
            spider="bga",
            item="RatingItem",
            full=full,
            keys=("bga_user_id", "bga_id"),
            fieldnames_exclude=("bgg_user_play_count",)
            if parse_bool(full)
            else ("bgg_user_play_count", "published_at", "updated_at", "scraped_at"),
        )
        return

    if spider == "bgg":
        yield merge_config(spider="bgg", item="GameItem", full=full)
        yield merge_config(
            spider="bgg",
            item="UserItem",
            full=full,
            keys="bgg_user_name",
            key_types="istr",
            fieldnames_exclude=None if full else ("published_at", "scraped_at"),
        )
        yield merge_config(
            spider="bgg",
            item="RatingItem",
            full=full,
            keys=("bgg_user_name", "bgg_id"),
            key_types=("istr", "int"),
            fieldnames_exclude=None if full else ("published_at", "scraped_at"),
        )
        return

    if spider == "bgg_hotness":
        yield merge_config(
            spider="bgg_hotness",
            item="GameItem",
            full=full,
            keys=("published_at", "bgg_id"),
            key_types=("date", "int"),
            latest_min=None,
            fieldnames=None
            if full
            else (
                "published_at",
                "rank",
                "add_rank",
                "bgg_id",
                "name",
                "year",
                "image_url",
            ),
            fieldnames_exclude=None,
            sort_keys=False,
            sort_fields=("published_at", "rank"),
        )
        return

    if spider == "bgg_rankings":
        yield merge_config(
            spider="bgg_rankings",
            item="GameItem",
            full=full,
            keys=("published_at", "bgg_id"),
            key_types=("date", "int"),
            latest_min=now() - timedelta(days=7),
            fieldnames=None
            if full
            else (
                "published_at",
                "bgg_id",
                "rank",
                "add_rank",
                "name",
                "year",
                "num_votes",
                "bayes_rating",
                "avg_rating",
            ),
            fieldnames_exclude=None,
            sort_keys=False,
            sort_fields=("published_at", "rank"),
        )
        return

    # TODO news merge config

    yield merge_config(spider=spider, item="GameItem", full=full)


def _parse_timeout(timeout):
    if timeout is None or timeout == "":
        return None

    timeout_float = parse_float(timeout)
    if timeout_float is not None:
        return timeout_float

    try:
        import pytimeparse
    except ImportError:
        return None

    return pytimeparse.parse(timeout)


def _docker_container(name):
    try:
        import docker
    except ImportError:
        LOGGER.warning("Docker library is not importable")
        return None

    try:
        client = docker.from_env()
        return client.containers.get(name)
    except docker.errors.NotFound:
        LOGGER.warning("Did not find container <%s>", name)

    return None


def _docker_start(name):
    LOGGER.info("Starting container <%s>", name)

    container = _docker_container(name)

    if container is not None:
        try:
            container.start()
            LOGGER.info("Started via Docker library call")
            return True
        except Exception:
            pass

    try:
        subprocess.run(["docker-compose", "start", name], check=True)
        LOGGER.info("Started via docker-compose CLI call")
        return True
    except Exception:
        pass

    LOGGER.warning("Unable to start container <%s>", name)
    return False


def _docker_stop(name, timeout=None):
    LOGGER.info("Stopping container <%s>", name)
    if timeout is not None:
        LOGGER.info("Allowing a timeout of %.1f seconds", timeout)

    container = _docker_container(name)

    if container is not None:
        try:
            if timeout is None:
                container.stop()
            else:
                container.stop(timeout=timeout)
            LOGGER.info("Stopped via Docker library call")
            return True
        except Exception:
            pass

    try:
        args = (
            ["docker-compose", "stop", name]
            if timeout is None
            else ["docker-compose", "stop", "--timeout", str(timeout), name]
        )
        subprocess.run(args=args, check=True)
        LOGGER.info("Stopped via docker-compose CLI call")
        return True
    except Exception:
        pass

    LOGGER.warning("Unable to stop container <%s>", name)
    return False


def _docker_compose(path, service):
    path = Path(path).resolve()
    LOGGER.info("Loading service <%s> from file <%s>", service, path)
    try:
        with open(path) as compose_file:
            config = safe_load(compose_file)
        return config["services"][service]
    except Exception:
        LOGGER.exception("Unable to load service <%s> from file <%s>", service, path)
    return {}


def _stop_merge_start(spider, compose_file, full=True, timeout=None, cool_down=None):
    scraper_name = spider.translate({ord("-"): "_"})
    docker_name = spider.translate({ord("_"): "-"})
    LOGGER.info(
        "Stopping, merging, and restarting spider <%s> / <%s>",
        scraper_name,
        docker_name,
    )

    docker_config = _docker_compose(path=compose_file, service=docker_name)
    container = docker_config.get("container_name")

    if not container:
        LOGGER.error("Unable to find container name for spider <%s>, aborting", spider)
        return False

    timeout = _parse_timeout(timeout)
    if timeout is None:
        timeout = _parse_timeout(docker_config.get("stop_grace_period"))

    # TODO add force option
    # if not _docker_stop(name=container, timeout=timeout):
    #     LOGGER.error("Unable to stop <%s>, aborting", container)
    #     return False

    _docker_stop(name=container, timeout=timeout)

    if cool_down:
        LOGGER.info("Cooling down for %d seconds...", cool_down)
        sleep(cool_down)

    for config in merge_configs(spider=scraper_name, full=full):
        LOGGER.info("Running merge with config %r", config)
        merge_files(**config)

    if cool_down:
        LOGGER.info("Cooling down for %d seconds...", cool_down)
        sleep(cool_down)

    return _docker_start(name=container)


def _parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("spiders", nargs="+", help="")
    parser.add_argument(
        "--compose-file", "-c", default=BASE_DIR / "docker-compose.yaml"
    )
    parser.add_argument("--timeout", "-t", help="")
    parser.add_argument("--cool-down", "-d", type=int, default=60, help="")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    for spider in args.spiders:
        try:
            _stop_merge_start(
                spider=spider,
                compose_file=args.compose_file,
                timeout=args.timeout,
                cool_down=args.cool_down,
            )
        except Exception:
            LOGGER.exception("There was an error when processing spider <%s>", spider)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

""" Spielen.de spider """

import os
import re

from pytility import clear_list, parse_int
from scrapy import Spider

from ..items import GameItem
from ..loaders import GameLoader
from ..utils import extract_spielen_id, now


def _parse_interval(text):
    match = re.match(r"^.*?(\d+)(\s*-\s*(\d+))?.*$", text)
    if match:
        return match.group(1), match.group(3)
    return None, None


def _parse_int(text):
    match = re.match(r"^.*?(\d+).*$", text)
    if match:
        return match.group(1)
    return None


class SpielenSpider(Spider):
    """Spielen.de spider"""

    name = "spielen"
    allowed_domains = ("spielen.de",)
    start_urls = (
        "https://gesellschaftsspiele.spielen.de/alle-brettspiele/",
        "https://gesellschaftsspiele.spielen.de/alle-brettspiele/?s=empfehlung",
        "https://gesellschaftsspiele.spielen.de/alle-brettspiele/?s=neue",
        "https://gesellschaftsspiele.spielen.de/alle-brettspiele/?s=name",
        "https://gesellschaftsspiele.spielen.de/messeneuheiten/",
    ) + tuple(
        f"https://gesellschaftsspiele.spielen.de/ausgezeichnet-{year}/"
        for year in range(2017, now().year + 1)
    )
    item_classes = (GameItem,)
    game_url = "https://gesellschaftsspiele.spielen.de/alle-brettspiele/{}/"

    custom_settings = {
        "DOWNLOAD_DELAY": 10,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 2,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 1,
        "LIMIT_IMAGES_TO_DOWNLOAD": parse_int(
            os.getenv("LIMIT_IMAGES_TO_DOWNLOAD_SPIELEN")
        )
        or 0,
    }

    def parse(self, response):
        """
        @url https://gesellschaftsspiele.spielen.de/alle-brettspiele/
        @returns items 0 0
        @returns requests 23
        """

        pagination = response.css(".listPagination a::attr(href)").extract()
        for page in clear_list(pagination):
            yield response.follow(page, callback=self.parse, priority=1)

        urls = response.xpath("//@href").extract()
        spielen_ids = map(extract_spielen_id, map(response.urljoin, urls))

        for spielen_id in clear_list(spielen_ids):
            yield response.follow(
                self.game_url.format(spielen_id),
                callback=self.parse_game,
                meta={"spielen_id": spielen_id},
            )

    # pylint: disable=no-self-use
    def parse_game(self, response):
        """
        @url https://gesellschaftsspiele.spielen.de/alle-brettspiele/catan-das-spiel/
        @returns items 1 1
        @returns requests 0 0
        @scrapes name year description designer artist publisher \
                 url image_url video_url rules_url \
                 min_players max_players min_age min_time max_time family \
                 num_votes avg_rating worst_rating best_rating \
                 complexity easiest_complexity hardest_complexity spielen_id
        """

        game = response.css("div.fullBox")

        ldr = GameLoader(
            item=GameItem(
                worst_rating=1,
                best_rating=5,
                easiest_complexity=1,
                hardest_complexity=5,
            ),
            selector=game,
            response=response,
        )

        spielen_id = response.meta.get("spielen_id") or extract_spielen_id(response.url)

        ldr.add_css("name", "h2")
        ldr.add_xpath(
            "year", './/div[b = "Erscheinungsjahr:"]/following-sibling::div//text()'
        )
        ldr.add_xpath("description", ".//h2/following-sibling::text()")

        ldr.add_xpath(
            "designer",
            './/div[b = "Autor:" or b = "Autoren:"]/following-sibling::div//text()',
        )
        ldr.add_xpath(
            "artist",
            './/div[b = "Illustrator:" or b = "Illustratoren:"]/following-sibling::div//text()',
        )
        ldr.add_xpath(
            "publisher",
            './/div[b = "Verlag:" or b = "Verlage:"]/following-sibling::div//a',
        )

        ldr.add_value("url", self.game_url.format(spielen_id))
        ldr.add_value("url", response.url)

        images = [
            game.xpath("(.//img)[1]/@data-src").extract_first(),
            game.xpath("(.//a[img])[1]/@href").extract_first(),
        ] + game.css("div.screenshotlist img::attr(data-large-src)").extract()
        ldr.add_value("image_url", (response.urljoin(i) for i in images if i))

        videos = (
            game.css("iframe::attr(src)").extract()
            + game.css("iframe::attr(data-src)").extract()
        )
        ldr.add_value("video_url", (response.urljoin(v) for v in videos if v))

        rules = game.xpath(
            './/a[@title = "Klicken zum Herunterladen."]/@href'
        ).extract()
        ldr.add_value("rules_url", map(response.urljoin, rules))

        players = game.xpath(
            './/div[b = "Spieler:"]/following-sibling::div/text()'
        ).extract_first()
        # TODO parse 'besonders gut mit 4 Spielern'
        min_players, max_players = _parse_interval(players) if players else (None, None)
        ldr.add_value("min_players", min_players)
        ldr.add_value("max_players", max_players)

        age = game.xpath(
            './/div[b = "Alter:"]/following-sibling::div/text()'
        ).extract_first()
        ldr.add_value("min_age", _parse_int(age) if age else None)

        time = game.xpath(
            './/div[b = "Dauer:"]/following-sibling::div/text()'
        ).extract_first()
        min_time, max_time = _parse_interval(time) if time else (None, None)
        ldr.add_value("min_time", min_time)
        ldr.add_value("max_time", max_time)

        ldr.add_xpath(
            "family",
            './/div[b = "Spielfamilie:" or b = "Spielfamilien:"]/following-sibling::div//text()',
        )

        ldr.add_css("num_votes", "span.votes")
        ldr.add_css("avg_rating", "span.average")

        complexity = game.xpath(
            './/div[. = "Komplexität:"]/following-sibling::div'
            '/span[following-sibling::span[contains(@class, "red")]]'
        )
        complexity = len(complexity) + 1
        ldr.add_value("complexity", complexity)

        ldr.add_value("spielen_id", spielen_id)

        return ldr.load_item()



# -*- coding: utf-8 -*-

""" DBpedia spider """

import os

from urllib.parse import urlencode

from pytility import batchify, parse_int, normalize_space
from scrapy import Request, Spider
from scrapy.utils.misc import arg_to_iter

from .wikidata import WikidataSpider
from ..items import GameItem
from ..loaders import GameLoader
from ..utils import extract_ids


def _sparql_xpath(
    prop,
    *,
    sparql_ns="s:",
    rooted=False,
    prop_var="property",
    value_var="value",
    value_type="literal",
    lang=None,
):
    sparql_ns = sparql_ns or ""
    sparql_ns = (
        sparql_ns + ":" if sparql_ns and not sparql_ns.endswith(":") else sparql_ns
    )
    root = "/{ns}sparql/{ns}results/" if rooted else ""
    result = '{ns}result[{ns}binding[@name = "{pv}"]/{ns}uri = "{prop}"]'
    binding = '/{ns}binding[@name = "{vv}"]/{ns}{vt}'
    lang_filter = '[@xml:lang = "{lang}"]' if lang else ""
    xpath = root + result + binding + lang_filter + "/text()"
    return xpath.format(
        ns=sparql_ns, pv=prop_var, prop=prop, vv=value_var, vt=value_type, lang=lang
    )


class DBpediaSpider(Spider):
    """DBpedia spider"""

    name = "dbpedia"
    allowed_domains = ("dbpedia.org",)
    item_classes = (GameItem,)
    sparql_api_url = "http://dbpedia.org/sparql"

    custom_settings = {
        "DOWNLOAD_DELAY": 20,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 4,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 2,
        "LIMIT_IMAGES_TO_DOWNLOAD": parse_int(
            os.getenv("LIMIT_IMAGES_TO_DOWNLOAD_DBPEDIA")
        )
        or 0,
    }

    game_types = (
        "http://dbpedia.org/ontology/BoardGame",
        "http://dbpedia.org/ontology/CardGame",
        # 'http://dbpedia.org/ontology/Game',
        "http://dbpedia.org/class/yago/BoardGame100502415",
        "http://dbpedia.org/class/yago/CardGame100488225",
        # 'http://dbpedia.org/class/yago/Game100455599',
        "http://dbpedia.org/class/yago/ParlorGame100501722",
        # 'http://dbpedia.org/class/yago/PartyGame100458800',
        "http://dbpedia.org/class/yago/Wikicat19th-centuryBoardGames",
        "http://dbpedia.org/class/yago/Wikicat19th-centuryCardGames",
        # 'http://dbpedia.org/class/yago/Wikicat3MBookshelfGames',
        "http://dbpedia.org/class/yago/WikicatAbstractStrategyGames",
        "http://dbpedia.org/class/yago/WikicatAdventureBoardGames",
        "http://dbpedia.org/class/yago/WikicatAlanR.MoonGames",
        "http://dbpedia.org/class/yago/WikicatAlderacEntertainmentGroupGames",
        "http://dbpedia.org/class/yago/WikicatAleaGames",
        "http://dbpedia.org/class/yago/WikicatAlexRandolphGames",
        # 'http://dbpedia.org/class/yago/WikicatAlternateHistoryGames',
        "http://dbpedia.org/class/yago/WikicatAmericanBoardGames",
        "http://dbpedia.org/class/yago/WikicatAmigoSpieleGames",
        "http://dbpedia.org/class/yago/WikicatAndreasSeyfarthGames",
        # 'http://dbpedia.org/class/yago/WikicatAndrewLooneyGames',
        # 'http://dbpedia.org/class/yago/WikicatAtlasGamesGames',
        "http://dbpedia.org/class/yago/WikicatAuctionBoardGames",
        "http://dbpedia.org/class/yago/WikicatAvalonHillGames",
        "http://dbpedia.org/class/yago/WikicatBiology-themedBoardGames",
        "http://dbpedia.org/class/yago/WikicatBoardGames",
        "http://dbpedia.org/class/yago/WikicatBoardGamesBasedOnMiddle-earth",
        "http://dbpedia.org/class/yago/WikicatBoardGamesBasedOnTelevisionSeries",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1928",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1934",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1947",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1951",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1952",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1955",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1957",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1958",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1960",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1962",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1965",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1967",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1970",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1971",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1974",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1975",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1976",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1977",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1978",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1980",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1982",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1983",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1984",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1985",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1986",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1987",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1988",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1989",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1990",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1991",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1992",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1994",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1995",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1996",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1997",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn1999",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2000",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2001",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2002",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2003",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2004",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2005",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2006",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2007",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2008",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2009",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2010",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedIn2011",
        "http://dbpedia.org/class/yago/WikicatBoardGamesIntroducedInThe1960s",
        "http://dbpedia.org/class/yago/WikicatBoardGamesUsingActionPoints",
        "http://dbpedia.org/class/yago/WikicatBoardGamesWithAModularBoard",
        "http://dbpedia.org/class/yago/WikicatBritishBoardGames",
        "http://dbpedia.org/class/yago/WikicatBrunoFaiduttiGames",
        "http://dbpedia.org/class/yago/WikicatBusinessSimulationGames",
        "http://dbpedia.org/class/yago/WikicatCardGames",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1972",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1979",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1982",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1992",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1993",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1994",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1995",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1996",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1997",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1998",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn1999",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2000",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2001",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2002",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2003",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2004",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2005",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2006",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2007",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2008",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2009",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2011",
        "http://dbpedia.org/class/yago/WikicatCardGamesIntroducedIn2012",
        # 'http://dbpedia.org/class/yago/WikicatChaosiumGames',
        # 'http://dbpedia.org/class/yago/WikicatCheapassGamesGames',
        "http://dbpedia.org/class/yago/WikicatChildren'sBoardGames",
        "http://dbpedia.org/class/yago/WikicatChineseAncientGames",
        # 'http://dbpedia.org/class/yago/WikicatChineseGames',
        # 'http://dbpedia.org/class/yago/WikicatClixGames',
        "http://dbpedia.org/class/yago/WikicatCollectible-basedGames",
        "http://dbpedia.org/class/yago/WikicatCollectibleActionFigureGames",
        "http://dbpedia.org/class/yago/WikicatCollectibleCardGames",
        "http://dbpedia.org/class/yago/WikicatCollectibleCardGamesBasedOnComics",
        "http://dbpedia.org/class/yago/WikicatCollectibleCardGamesBasedOnMarvelComics",
        "http://dbpedia.org/class/yago/WikicatCollectibleMiniaturesGames",
        # 'http://dbpedia.org/class/yago/WikicatColumbiaGamesGames',
        # 'http://dbpedia.org/class/yago/WikicatConnectionGames',
        "http://dbpedia.org/class/yago/WikicatCooperativeBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatCthulhuMythosGames',
        "http://dbpedia.org/class/yago/WikicatCzechBoardGames",
        "http://dbpedia.org/class/yago/WikicatCzechGamesEditionGames",
        "http://dbpedia.org/class/yago/WikicatDaysOfWonderGames",
        "http://dbpedia.org/class/yago/WikicatDedicatedDeckCardGames",
        "http://dbpedia.org/class/yago/WikicatDeductionBoardGames",
        "http://dbpedia.org/class/yago/WikicatDiceGames",
        # 'http://dbpedia.org/class/yago/WikicatDonGreenwoodGames',
        # 'http://dbpedia.org/class/yago/WikicatEagleGamesGames',
        "http://dbpedia.org/class/yago/WikicatEconomicSimulationBoardGames",
        "http://dbpedia.org/class/yago/WikicatEducationalBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatEthicsGames',
        "http://dbpedia.org/class/yago/WikicatFantasyBoardGames",
        "http://dbpedia.org/class/yago/WikicatFantasyFlightGamesGames",
        # 'http://dbpedia.org/class/yago/WikicatFantasyGames',
        # 'http://dbpedia.org/class/yago/WikicatFinnishGames',
        # 'http://dbpedia.org/class/yago/WikicatFiveRingsPublishingGroupGames',
        # 'http://dbpedia.org/class/yago/WikicatFrankChadwickGames',
        "http://dbpedia.org/class/yago/WikicatGMTGamesGames",
        "http://dbpedia.org/class/yago/WikicatGameDesigners'WorkshopGames",
        # 'http://dbpedia.org/class/yago/WikicatGamesBasedOnDune',
        # 'http://dbpedia.org/class/yago/WikicatGamesBasedOnMiddle-earth',
        # 'http://dbpedia.org/class/yago/WikicatGamesBasedOnStarTrek',
        "http://dbpedia.org/class/yago/WikicatGamesOfMentalSkill",
        # 'http://dbpedia.org/class/yago/WikicatGamesOfPhysicalSkill',
        "http://dbpedia.org/class/yago/WikicatGamesWorkshopGames",
        # 'http://dbpedia.org/class/yago/WikicatGaryGygaxGames',
        "http://dbpedia.org/class/yago/WikicatHansImGlückGames",
        "http://dbpedia.org/class/yago/WikicatHasbroGames",
        "http://dbpedia.org/class/yago/WikicatHistoricalBoardGames",
        "http://dbpedia.org/class/yago/WikicatHorrorBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatIcehouseGames',
        # 'http://dbpedia.org/class/yago/WikicatIronCrownEnterprisesGames',
        # 'http://dbpedia.org/class/yago/WikicatJimDunniganGames',
        # 'http://dbpedia.org/class/yago/WikicatJonathanTweetGames',
        "http://dbpedia.org/class/yago/WikicatKlausTeuberGames",
        # 'http://dbpedia.org/class/yago/WikicatKoreanGames',
        "http://dbpedia.org/class/yago/WikicatKosmos(publisher)Games",
        "http://dbpedia.org/class/yago/WikicatKrisBurmGames",
        # 'http://dbpedia.org/class/yago/WikicatLarryHarrisGames',
        "http://dbpedia.org/class/yago/WikicatLicensedBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatMarcMillerGames',
        # 'http://dbpedia.org/class/yago/WikicatMarkRein·HagenGames',
        "http://dbpedia.org/class/yago/WikicatMartinWallaceGames",
        "http://dbpedia.org/class/yago/WikicatMatchingCardGames",
        "http://dbpedia.org/class/yago/WikicatMattelGames",
        "http://dbpedia.org/class/yago/WikicatMayfairGamesGames",
        "http://dbpedia.org/class/yago/WikicatMichaelSchachtGames",
        # 'http://dbpedia.org/class/yago/WikicatMikeFitzgeraldGames',
        # 'http://dbpedia.org/class/yago/WikicatMiltonBradleyCompanyGames',
        "http://dbpedia.org/class/yago/WikicatMiniaturesGames",
        # 'http://dbpedia.org/class/yago/WikicatMultiplayerGames',
        # 'http://dbpedia.org/class/yago/WikicatNavalGames',
        "http://dbpedia.org/class/yago/WikicatNegotiationTabletopGames",
        "http://dbpedia.org/class/yago/WikicatPaper-and-pencilGames",
        "http://dbpedia.org/class/yago/WikicatParkerBrothersGames",
        "http://dbpedia.org/class/yago/WikicatPartyBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatPartyGames',
        # 'http://dbpedia.org/class/yago/WikicatPinnacleEntertainmentGroupGames',
        "http://dbpedia.org/class/yago/WikicatPlayingCards",
        "http://dbpedia.org/class/yago/WikicatQueenGamesGames",
        # 'http://dbpedia.org/class/yago/WikicatQuizGames',
        # 'http://dbpedia.org/class/yago/WikicatRaceGames',
        "http://dbpedia.org/class/yago/WikicatRacingBoardGames",
        "http://dbpedia.org/class/yago/WikicatRailroadBoardGames",
        "http://dbpedia.org/class/yago/WikicatRavensburgerGames",
        "http://dbpedia.org/class/yago/WikicatReinerKniziaGames",
        "http://dbpedia.org/class/yago/WikicatRichardBergGames",
        "http://dbpedia.org/class/yago/WikicatRichardBorgGames",
        "http://dbpedia.org/class/yago/WikicatRichardGarfieldGames",
        "http://dbpedia.org/class/yago/WikicatRioGrandeGamesGames",
        "http://dbpedia.org/class/yago/WikicatRole-playingGames",
        "http://dbpedia.org/class/yago/WikicatRoll-and-moveBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatSchaperToysGames',
        "http://dbpedia.org/class/yago/WikicatScienceFictionBoardGames",
        "http://dbpedia.org/class/yago/WikicatShedding-typeCardGames",
        "http://dbpedia.org/class/yago/WikicatSidSacksonGames",
        # 'http://dbpedia.org/class/yago/WikicatSingle-playerGames',
        "http://dbpedia.org/class/yago/WikicatSportsBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatStarWarsGames',
        # 'http://dbpedia.org/class/yago/WikicatSteampunkGames',
        "http://dbpedia.org/class/yago/WikicatSteveJacksonGamesGames",
        # 'http://dbpedia.org/class/yago/WikicatTSR(company)Games',
        "http://dbpedia.org/class/yago/WikicatTabletopGames",
        "http://dbpedia.org/class/yago/WikicatTile-basedBoardGames",
        "http://dbpedia.org/class/yago/WikicatTile-layingBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatTomWhamGames',
        "http://dbpedia.org/class/yago/WikicatTraditionalBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatUpperDeckCompanyGames',
        "http://dbpedia.org/class/yago/WikicatUweRosenbergGames",
        "http://dbpedia.org/class/yago/WikicatWarhammer40,000CardGames",
        "http://dbpedia.org/class/yago/WikicatWarhammer40,000TabletopGames",
        "http://dbpedia.org/class/yago/WikicatWinningMovesGames",
        # 'http://dbpedia.org/class/yago/WikicatWizKidsGames',
        "http://dbpedia.org/class/yago/WikicatWolfgangKramerGames",
        "http://dbpedia.org/class/yago/WikicatWordBoardGames",
        # 'http://dbpedia.org/class/yago/WikicatWordGames',
        "http://dbpedia.org/class/yago/WikicatZ-ManGamesGames",
        # 'http://dbpedia.org/class/yago/WordGame100501870',
    )

    def _api_url(self, query):
        args = {"format": "text/xml", "query": query}
        return "{}?{}".format(self.sparql_api_url, urlencode(args))

    # def _entity_url(self, wikidata_id, fformat='json'):
    #     return self.entity_data_url.format(wikidata_id=wikidata_id, fformat=fformat)

    def _type_requests(self, types, batch_size=10):
        query_tmpl = (
            "SELECT DISTINCT ?game WHERE {{ ?game a ?type . VALUES ?type {{ {} }} }}"
        )

        for batch in batchify(types, batch_size):
            query = query_tmpl.format(" ".join(batch))
            # self.logger.debug(query)
            yield Request(self._api_url(query), callback=self.parse_games, priority=1)

    def start_requests(self):
        """generate start requests"""

        dbp_types = getattr(self, "game_types", None) or ()
        wd_types = getattr(WikidataSpider, "game_types", None) or ()
        types = tuple("<{}>".format(t) for t in dbp_types) + tuple(
            "<http://www.wikidata.org/entity/{}>".format(t) for t in wd_types
        )

        if types:
            yield from self._type_requests(types)
            return

        query = normalize_space(
            """
            SELECT DISTINCT ?type WHERE {
                ?game <http://dbpedia.org/property/bggid> ?bgg;
                      a ?type .
            }"""
        )
        # self.logger.debug(query)
        yield Request(self._api_url(query), callback=self.parse, priority=2)

    def parse(self, response):
        # pylint: disable=line-too-long
        """
        @url http://dbpedia.org/sparql?query=SELECT%20DISTINCT%20%3Ftype%20WHERE%20%7B%20%3Fgame%20%3Chttp%3A//dbpedia.org/property/bggid%3E%20%3Fbgg%3B%20a%20%3Ftype%20.%20%7D&format=text%2Fxml
        @returns items 0 0
        @returns requests 40
        """

        response.selector.register_namespace(
            "s", "http://www.w3.org/2005/sparql-results#"
        )
        types = response.xpath('//s:binding[@name = "type"]/s:uri/text()').extract()
        self.logger.info("received %d types", len(types))
        yield from self._type_requests(f"<{t}>" for t in types)

    def parse_games(self, response):
        # pylint: disable=line-too-long
        """
        @url http://dbpedia.org/sparql?query=SELECT+DISTINCT+%3Fgame+WHERE+%7B+%3Fgame+a+%3Chttp%3A%2F%2Fdbpedia.org%2Fclass%2Fyago%2FBoardGame100502415%3E+.+%7D&format=text%2Fxml
        @returns items 0 0
        @returns requests 1200
        """

        response.selector.register_namespace(
            "s", "http://www.w3.org/2005/sparql-results#"
        )

        games = response.xpath('//s:binding[@name = "game"]/s:uri/text()').extract()

        self.logger.info("received %d games", len(games))

        query_tmpl = normalize_space(
            """
            SELECT ?property ?value ?label WHERE {{
                <{game}> ?property ?value .
                OPTIONAL {{ ?value <http://www.w3.org/2000/01/rdf-schema#label> ?label . }}
            }}"""
        )

        for game in games:
            # dbpedia_id = game.split('/')[-1]
            # http://dbpedia.org/resource/{dbpedia_id}
            query = query_tmpl.format(game=game)
            # self.logger.debug(query)
            yield Request(
                self._api_url(query),
                callback=self.parse_game,
                meta={"dbpedia_uri": game},
            )

    def parse_game(self, response):
        # pylint: disable=line-too-long
        """
        @url http://dbpedia.org/sparql?query=SELECT+%3Fproperty+%3Fvalue+%3Flabel+WHERE+%7B+%3Chttp%3A%2F%2Fdbpedia.org%2Fresource%2FCatan%3E+%3Fproperty+%3Fvalue+.+OPTIONAL+%7B+%3Fvalue+%3Chttp%3A%2F%2Fwww.w3.org%2F2000%2F01%2Frdf-schema%23label%3E+%3Flabel+.+%7D+%7D&format=text%2Fxml
        @returns items 1 1
        @returns requests 0 0
        @scrapes name alt_name year description designer publisher \
            official_url image_url external_link \
            min_players min_age bgg_id freebase_id wikidata_id wikipedia_id dbpedia_id
        """

        response.selector.register_namespace(
            "xml", "http://www.w3.org/XML/1998/namespace"
        )
        response.selector.register_namespace(
            "s", "http://www.w3.org/2005/sparql-results#"
        )

        results = response.xpath("/s:sparql/s:results")

        if not results or len(results) != 1:
            self.logger.warning("no or unexpected results found in %r", response)
            return None

        uri = response.meta.get("dbpedia_uri")

        ldr = GameLoader(item=GameItem(), selector=results[0], response=response)

        #  'http://dbpedia.org/property/id', # seems useless
        #  'http://dbpedia.org/property/playingTime',
        #  'http://dbpedia.org/property/randomChance',
        #  'http://dbpedia.org/property/setupTime',
        #  'http://dbpedia.org/property/skills',
        #  'http://dbpedia.org/property/title', # awards

        ldr.add_xpath(
            "name",
            _sparql_xpath("http://www.w3.org/2000/01/rdf-schema#label", lang="en"),
        )
        ldr.add_xpath(
            "name", _sparql_xpath("http://xmlns.com/foaf/0.1/name", lang="en")
        )
        ldr.add_xpath(
            "name", _sparql_xpath("http://dbpedia.org/property/name", lang="en")
        )
        ldr.add_xpath(
            "name", _sparql_xpath("http://www.w3.org/2000/01/rdf-schema#label")
        )
        ldr.add_xpath("name", _sparql_xpath("http://xmlns.com/foaf/0.1/name"))
        ldr.add_xpath("name", _sparql_xpath("http://dbpedia.org/property/name"))
        ldr.add_xpath(
            "alt_name", _sparql_xpath("http://www.w3.org/2000/01/rdf-schema#label")
        )
        ldr.add_xpath("alt_name", _sparql_xpath("http://xmlns.com/foaf/0.1/name"))
        ldr.add_xpath("alt_name", _sparql_xpath("http://dbpedia.org/property/name"))
        ldr.add_xpath("year", _sparql_xpath("http://dbpedia.org/property/date"))
        ldr.add_xpath("year", _sparql_xpath("http://dbpedia.org/property/years"))
        ldr.add_xpath(
            "description",
            _sparql_xpath("http://dbpedia.org/ontology/abstract", lang="en"),
        )
        ldr.add_xpath(
            "description",
            _sparql_xpath("http://www.w3.org/2000/01/rdf-schema#comment", lang="en"),
        )
        ldr.add_xpath(
            "description", _sparql_xpath("http://dbpedia.org/ontology/abstract")
        )
        ldr.add_xpath(
            "description", _sparql_xpath("http://www.w3.org/2000/01/rdf-schema#comment")
        )

        ldr.add_xpath(
            "designer",
            _sparql_xpath(
                "http://dbpedia.org/ontology/designer", value_var="label", lang="en"
            ),
        )
        ldr.add_xpath(
            "designer",
            _sparql_xpath("http://dbpedia.org/ontology/designer", value_var="label"),
        )
        ldr.add_xpath(
            "publisher",
            _sparql_xpath(
                "http://dbpedia.org/ontology/publisher", value_var="label", lang="en"
            ),
        )
        ldr.add_xpath(
            "publisher",
            _sparql_xpath("http://dbpedia.org/ontology/publisher", value_var="label"),
        )

        ldr.add_value("url", uri)
        ldr.add_xpath(
            "official_url",
            _sparql_xpath("http://xmlns.com/foaf/0.1/homepage", value_type="uri"),
        )
        ldr.add_xpath(
            "official_url",
            _sparql_xpath("http://dbpedia.org/property/web", value_type="uri"),
        )
        ldr.add_xpath(
            "image_url",
            _sparql_xpath("http://xmlns.com/foaf/0.1/depiction", value_type="uri"),
        )
        ldr.add_xpath(
            "image_url",
            _sparql_xpath("http://dbpedia.org/ontology/thumbnail", value_type="uri"),
        )
        ldr.add_xpath(
            "image_url",
            _sparql_xpath("http://dbpedia.org/property/imageLink", value_type="uri"),
        )
        ldr.add_xpath(
            "external_link",
            _sparql_xpath(
                "http://dbpedia.org/ontology/wikiPageExternalLink", value_type="uri"
            ),
        )
        ldr.add_xpath(
            "external_link",
            _sparql_xpath(
                "http://xmlns.com/foaf/0.1/isPrimaryTopicOf", value_type="uri"
            ),
        )
        ldr.add_xpath(
            "external_link",
            _sparql_xpath("http://www.w3.org/2002/07/owl#sameAs", value_type="uri"),
        )

        ldr.add_xpath(
            "min_players", _sparql_xpath("http://dbpedia.org/property/players")
        )
        ldr.add_xpath("min_age", _sparql_xpath("http://dbpedia.org/property/ages"))

        ldr.add_xpath("bgg_id", _sparql_xpath("http://dbpedia.org/property/bggid"))
        ldr.add_value(
            None,
            extract_ids(
                uri,
                *arg_to_iter(ldr.get_output_value("external_link")),
                *arg_to_iter(ldr.get_output_value("official_url")),
            ),
        )

        return ldr.load_item()



# -*- coding: utf-8 -*-

""" Board Game Atlas spider """

import os

from functools import partial
from itertools import chain
from urllib.parse import urlencode

from pytility import parse_float, parse_int
from scrapy import Request, Spider
from scrapy.utils.project import get_project_settings

from ..items import GameItem, RatingItem
from ..loaders import GameJsonLoader, RatingJsonLoader
from ..utils import (
    extract_bga_id,
    extract_meta,
    extract_item,
    extract_query_param,
    extract_url,
    json_from_response,
    now,
)

API_URL = "https://api.boardgameatlas.com/api"


def _extract_bga_id(item=None, response=None):
    if item and item.get("bga_id"):
        return item["bga_id"]
    meta = extract_meta(response)
    if meta.get("bga_id"):
        return meta["bga_id"]
    url = extract_url(item, response)
    return extract_bga_id(url)


def _extract_requests(response=None):
    meta = extract_meta(response)
    return meta.get("game_requests")


class BgaSpider(Spider):
    """Board Game Atlas spider"""

    name = "bga"
    allowed_domains = ("boardgameatlas.com",)
    item_classes = (GameItem, RatingItem)
    api_url = API_URL

    custom_settings = {
        "IMAGES_URLS_FIELD": None,
        "DOWNLOAD_DELAY": 30,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 4,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 2,
        "LIMIT_IMAGES_TO_DOWNLOAD": parse_int(os.getenv("LIMIT_IMAGES_TO_DOWNLOAD_BGA"))
        or 0,
    }

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        """initialise spider from crawler"""

        kwargs.setdefault("settings", crawler.settings)
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        return spider

    def __init__(self, *args, settings=None, **kwargs):
        super().__init__(*args, **kwargs)
        settings = settings or get_project_settings()
        self.client_id = settings.get("BGA_CLIENT_ID")
        self.scrape_images = settings.getbool("BGA_SCRAPE_IMAGES")
        self.scrape_videos = settings.getbool("BGA_SCRAPE_VIDEOS")
        self.scrape_reviews = settings.getbool("BGA_SCRAPE_REVIEWS")

    def _api_url(self, path="search", query=None):
        query = query or {}
        query.setdefault("client_id", self.client_id)
        query.setdefault("limit", 100)
        return "{}/{}?{}".format(
            self.api_url, path, urlencode(sorted(query.items(), key=lambda x: x[0]))
        )

    def _game_requests(self, bga_id):
        if self.scrape_images:
            yield self._api_url("game/images", {"game_id": bga_id}), self.parse_images
        if self.scrape_videos:
            yield self._api_url("game/videos", {"game_id": bga_id}), self.parse_videos
        if self.scrape_reviews:
            yield self._api_url("game/reviews", {"game_id": bga_id}), self.parse_reviews

    # pylint: disable=no-self-use
    def _next_request_or_item(self, item, requests):
        if not requests:
            return item

        url, callback = requests.pop(0)
        callback = partial(callback, item=item)
        return Request(
            url=url,
            callback=callback,
            errback=callback,
            meta={"item": item, "game_requests": requests},
        )

    def start_requests(self):
        """generate start requests"""

        yield Request(
            url=self._api_url(query={"order_by": "popularity"}),
            callback=self.parse,
            priority=2,
        )
        yield Request(
            url=self._api_url(path="reviews"),
            callback=self.parse_user_reviews,
            priority=1,
        )

    # pylint: disable=line-too-long
    def parse(self, response):
        """
        @url https://api.boardgameatlas.com/api/search?client_id=8jfqHypg2l&order_by=popularity&limit=100
        @returns items 100 100
        @returns requests 1 1
        @scrapes name description url image_url bga_id scraped_at worst_rating best_rating
        """

        result = json_from_response(response)
        games = result.get("games") or ()
        scraped_at = now()

        if games:
            skip = parse_int(extract_query_param(response.url, "skip")) or 0
            limit = parse_int(extract_query_param(response.url, "limit")) or 100
            query = {"order_by": "popularity", "skip": skip + limit, "limit": limit}
            yield Request(
                url=self._api_url(query=query), callback=self.parse, priority=2
            )

        for game in games:
            bga_id = game.get("id") or extract_bga_id(game.get("url"))
            ldr = GameJsonLoader(
                item=GameItem(
                    bga_id=bga_id, scraped_at=scraped_at, worst_rating=1, best_rating=5
                ),
                json_obj=game,
                response=response,
            )

            ldr.add_jmes("name", "name")
            ldr.add_jmes("alt_name", "names")
            ldr.add_jmes("year", "year_published")
            ldr.add_jmes("description", "description_preview")
            ldr.add_jmes("description", "description")

            ldr.add_jmes("designer", "designers")
            ldr.add_jmes("artist", "artists")
            ldr.add_jmes("publisher", "primary_publisher")
            ldr.add_jmes("publisher", "publishers")

            ldr.add_jmes("url", "url")
            ldr.add_jmes("image_url", "image_url")
            ldr.add_jmes("image_url", "thumb_url")
            ldr.add_jmes("rules_url", "rules_url")
            ldr.add_jmes("external_link", "official_url")

            list_price = ldr.get_jmes("msrp")
            list_price = map(
                "USD{:.2f}".format, filter(None, map(parse_float, list_price))
            )
            ldr.add_value("list_price", list_price)

            ldr.add_jmes("min_players", "min_players")
            ldr.add_jmes("max_players", "max_players")
            ldr.add_jmes("min_age", "min_age")
            ldr.add_jmes("min_time", "min_playtime")
            ldr.add_jmes("max_time", "max_playtime")

            # TODO resolve mechanic and category (#48)
            # https://www.boardgameatlas.com/api/docs/game/categories
            # https://www.boardgameatlas.com/api/docs/game/mechanics
            ldr.add_jmes("category", "categories[].id")
            ldr.add_jmes("mechanic", "mechanics[].id")

            ldr.add_jmes("num_votes", "num_user_ratings")
            ldr.add_jmes("avg_rating", "average_user_rating")

            item = ldr.load_item()
            requests = list(self._game_requests(bga_id))
            yield self._next_request_or_item(item, requests)

    def parse_images(self, response, item=None):
        """
        @url https://api.boardgameatlas.com/api/game/images?client_id=8jfqHypg2l&game_id=OIXt3DmJU0&limit=100
        @returns items 1 1
        @returns requests 0 0
        @scrapes image_url
        """

        item = extract_item(item, response, GameItem)
        result = json_from_response(response)

        ldr = GameJsonLoader(item=item, json_obj=result, response=response)
        ldr.add_value("image_url", item.get("image_url"))
        ldr.add_jmes("image_url", "images[].url")
        ldr.add_jmes("image_url", "images[].thumb")

        item = ldr.load_item()
        requests = _extract_requests(response)
        return self._next_request_or_item(item, requests)

    def parse_videos(self, response, item=None):
        """
        @url https://api.boardgameatlas.com/api/game/videos?client_id=8jfqHypg2l&game_id=OIXt3DmJU0&limit=100
        @returns items 1 1
        @returns requests 0 0
        @scrapes video_url
        """

        item = extract_item(item, response, GameItem)
        result = json_from_response(response)

        ldr = GameJsonLoader(item=item, json_obj=result, response=response)
        ldr.add_value("video_url", item.get("video_url"))
        ldr.add_jmes("video_url", "videos[].url")

        item = ldr.load_item()
        requests = _extract_requests(response)
        return self._next_request_or_item(item, requests)

    # pylint: disable=no-self-use
    def parse_reviews(self, response, item=None):
        """
        @url https://api.boardgameatlas.com/api/game/reviews?client_id=8jfqHypg2l&game_id=OIXt3DmJU0&limit=100
        @returns items 1 1
        @returns requests 0 0
        @scrapes review_url
        """

        item = extract_item(item, response, GameItem)
        result = json_from_response(response)

        ldr = GameJsonLoader(item=item, json_obj=result, response=response)
        ldr.add_value("review_url", item.get("review_url"))
        ldr.add_jmes("review_url", "reviews[].url")

        item = ldr.load_item()
        requests = _extract_requests(response)
        return self._next_request_or_item(item, requests)

    def parse_user_reviews(self, response):
        """
        @url https://api.boardgameatlas.com/api/reviews?client_id=8jfqHypg2l&limit=100
        @returns items 100 100
        @returns requests 1 1
        @scrapes item_id bga_id bga_user_id bga_user_name
        """

        result = json_from_response(response)
        reviews = result.get("reviews") or ()
        scraped_at = now()

        if reviews:
            skip = parse_int(extract_query_param(response.url, "skip")) or 0
            limit = parse_int(extract_query_param(response.url, "limit")) or 100
            query = {"skip": skip + limit, "limit": limit}
            yield Request(
                url=self._api_url(path="reviews", query=query),
                callback=self.parse_user_reviews,
                priority=1,
            )

        for review in reviews:
            ldr = RatingJsonLoader(
                item=RatingItem(scraped_at=scraped_at),
                json_obj=review,
                response=response,
            )

            ldr.add_jmes("item_id", "id")

            ldr.add_jmes("bga_id", "game.id.objectId")
            ldr.add_jmes("bga_user_id", "user.id")
            ldr.add_jmes("bga_user_name", "user.username")
            ldr.add_jmes("bga_user_rating", "rating")
            comments = chain(ldr.get_jmes("title"), ldr.get_jmes("description"))
            ldr.add_value("comment", "\n".join(filter(None, comments)))

            yield ldr.load_item()



# -*- coding: utf-8 -*-

""" Scrapy item pipelines """

import logging
import math
import re

from itertools import islice
from urllib.parse import quote, unquote_plus
from typing import Optional

import jmespath

from itemadapter import ItemAdapter
from pytility import clear_list, take_first
from scrapy import Request
from scrapy.exceptions import DropItem, NotConfigured
from scrapy.utils.defer import defer_result
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.python import flatten
from twisted.internet.defer import DeferredList

from .utils import REGEX_DBPEDIA_DOMAIN, parse_json, parse_url

LOGGER = logging.getLogger(__name__)


class DataTypePipeline:
    """convert fields to their required data type"""

    # pylint: disable=no-self-use,unused-argument
    def process_item(self, item, spider):
        """convert to data type"""

        for field in item.fields:
            dtype = item.fields[field].get("dtype")
            default = item.fields[field].get("default", NotImplemented)

            if item.get(field) is None and default is not NotImplemented:
                item[field] = default() if callable(default) else default

            if not dtype or item.get(field) is None or isinstance(item[field], dtype):
                continue

            try:
                item[field] = dtype(item[field])
            except Exception as exc:
                if default is NotImplemented:
                    raise DropItem(
                        'Could not convert field "{}" to datatype "{}" in item "{}"'.format(
                            field, dtype, item
                        )
                    ) from exc

                item[field] = default() if callable(default) else default

        return item


class ResolveLabelPipeline:
    """resolve labels"""

    @classmethod
    def from_crawler(cls, crawler):
        """init from crawler"""

        url = crawler.settings.get("RESOLVE_LABEL_URL")
        fields = crawler.settings.getlist("RESOLVE_LABEL_FIELDS")

        if not url or not fields:
            raise NotConfigured

        lang_priorities = crawler.settings.getlist("RESOLVE_LABEL_LANGUAGE_PRIORITIES")

        return cls(url=url, fields=fields, lang_priorities=lang_priorities)

    def __init__(self, url, fields, lang_priorities=None):
        self.url = url
        self.fields = fields
        self.lang_priorities = {
            lang: prio for prio, lang in enumerate(arg_to_iter(lang_priorities))
        }
        self.labels = {}
        self.logger = LOGGER

    def _extract_labels(self, response, value):
        json_obj = parse_json(response.text) if hasattr(response, "text") else {}

        labels = take_first(jmespath.search(f"entities.{value}.labels", json_obj)) or {}
        labels = labels.values()
        labels = sorted(
            labels,
            key=lambda label: self.lang_priorities.get(label.get("language"), math.inf),
        )
        labels = clear_list(label.get("value") for label in labels)

        self.labels[value] = labels
        self.logger.debug("resolved labels for %s: %s", value, labels)

        return labels

    def _deferred_value(self, value, spider):
        labels = self.labels.get(value)
        if labels is not None:
            self.logger.debug("found labels in cache for %s: %s", value, labels)
            return defer_result(labels)

        request = Request(self.url.format(value), priority=1)
        deferred = spider.crawler.engine.download(request, spider)
        deferred.addBoth(self._extract_labels, value)
        return deferred

    def _add_value(self, result, field, item):
        labels = clear_list(flatten(r[1] for r in arg_to_iter(result))) or None
        self.logger.debug("resolved labels for %s: %s", item.get(field), labels)
        item[field] = labels
        return item

    def _deferred_field(self, field, item, spider):
        deferreds = [
            self._deferred_value(value, spider)
            for value in arg_to_iter(item.get(field))
        ]
        if not deferreds:
            item[field] = None
            return defer_result(item)
        deferred = DeferredList(deferreds, consumeErrors=True)
        deferred.addBoth(self._add_value, field, item)
        return deferred

    def process_item(self, item, spider):
        """resolve IDs to labels in specified fields"""

        if not any(item.get(field) for field in self.fields):
            return item

        deferred = DeferredList(
            [self._deferred_field(field, item, spider) for field in self.fields],
            consumeErrors=True,
        )
        deferred.addBoth(lambda _: item)
        return deferred


class ResolveImagePipeline:
    """resolve image URLs"""

    fields = ("image_url",)
    hostnames = (
        "dbpedia.org",
        "www.dbpedia.org",
        "wikidata.org",
        "www.wikidata.org",
        REGEX_DBPEDIA_DOMAIN,
    )
    regex_path = re.compile(r"^/(resource/File:|wiki/Special:EntityData/)(.+)$")
    url = "https://commons.wikimedia.org/wiki/Special:Redirect/file/{}"
    logger = LOGGER

    def _parse_url(self, url):
        parsed = parse_url(url, self.hostnames)
        if not parsed:
            return url

        match = self.regex_path.match(parsed.path)
        if not match:
            return url

        commons_id = unquote_plus(match.group(2))
        commons_id = commons_id.replace(" ", "_")
        commons_id = quote(commons_id)

        result = self.url.format(commons_id)
        self.logger.debug("converted URL <%s> to <%s>", url, result)
        return result

    # pylint: disable=unused-argument
    def process_item(self, item, spider):
        """resolve resource image URLs to actual file locations"""
        for field in self.fields:
            if item.get(field):
                item[field] = clear_list(map(self._parse_url, arg_to_iter(item[field])))
        return item


class LimitImagesPipeline:
    """Copy a limited number of image URLs to be downloaded from source to target."""

    source_field: str
    target_field: str
    limit: Optional[int] = None

    @classmethod
    def from_crawler(cls, crawler):
        """Init from crawler."""

        source_field = crawler.settings.get("LIMIT_IMAGES_URLS_FIELD")
        target_field = crawler.settings.get("IMAGES_URLS_FIELD")

        if not source_field or not target_field:
            raise NotConfigured

        limit = crawler.settings.getint("LIMIT_IMAGES_TO_DOWNLOAD", -1)

        return cls(
            source_field=source_field,
            target_field=target_field,
            limit=limit,
        )

    def __init__(
        self, source_field: str, target_field: str, limit: Optional[int] = None
    ):
        self.source_field = source_field
        self.target_field = target_field
        self.limit = limit

    # pylint: disable=unused-argument
    def process_item(self, item, spider):
        """Copy a limited number of image URLs to be downloaded from source to target."""

        # adding target field would result in error; return item as-is
        if hasattr(item, "fields") and self.target_field not in item.fields:
            return item

        if self.limit is None or self.limit < 0:  # copy through everything
            item[self.target_field] = list(arg_to_iter(item.get(self.source_field)))
            return item

        if not self.limit:  # limit is zero
            item[self.target_field] = []
            return item

        # actual limit
        item[self.target_field] = list(
            islice(arg_to_iter(item.get(self.source_field)), self.limit)
        )
        return item


class CleanItemPipeline:
    """Clean up unnecessary values from an item."""

    drop_falsey: bool
    drop_values: Optional[tuple]

    @classmethod
    def from_crawler(cls, crawler):
        """Init from crawler."""

        drop_falsey = crawler.settings.getbool("CLEAN_ITEM_DROP_FALSEY")
        drop_values = (
            tuple(arg_to_iter(crawler.settings.getlist("CLEAN_ITEM_DROP_VALUES")))
            or None
        )

        if not drop_falsey and not drop_values:
            raise NotConfigured

        return cls(drop_falsey=drop_falsey, drop_values=drop_values)

    def __init__(self, drop_falsey: bool, drop_values: Optional[tuple]):
        self.drop_falsey = drop_falsey
        self.drop_values = drop_values

    # pylint: disable=unused-argument
    def process_item(self, item, spider):
        """Clean up unnecessary values from an item."""

        adapter = ItemAdapter(item)

        for key in tuple(adapter.keys()):
            if (self.drop_falsey and not adapter[key]) or (
                self.drop_values is not None and adapter[key] in self.drop_values
            ):
                del adapter[key]

        return item
